# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Running Ray Tuners: https://docs.ray.io/en/latest/ray-air/tuner.html."""

from typing import Any, Callable, List, Optional, Tuple

import numpy as np
from ray import air
from ray import data
from ray import tune
from ray.air import session
from vizier import pyvizier
from vizier._src.raytune import converters
from vizier.benchmarks import experimenters


# See https://docs.ray.io/en/latest/data/dataset.html
def run_tune_distributed(
    run_tune_args_list: List[Tuple[Any]],
    run_tune: Callable[[Any], tune.result_grid.ResultGrid],
) -> List[tune.result_grid.ResultGrid]:
  """Distributes tuning via Ray datasets API for MapReduce purposes.

  NOTE: There are no datasets processed. However, all MapReduce operations
  are now done in the Datasets API under Ray.

  Args:
    run_tune_args_list: List of Tuples that are to be passed into run_tune.
    run_tune: Callable that accepts args from previous list.

  Returns:
    List of results.
  """
  ds = data.from_items([{'args_tuple': args} for args in run_tune_args_list])
  ds = ds.map(lambda x: {'result': run_tune(*x['args_tuple'])})
  return ds.take_all()


def run_tune_bbob(
    function_name: str,
    dimension: int,
    shift: Optional[np.ndarray] = None,
    tune_config: Optional[tune.TuneConfig] = None,
    run_config: Optional[air.RunConfig] = None,
) -> tune.result_grid.ResultGrid:
  """Runs Ray Tuners for BBOB problems.

  See https://docs.ray.io/en/latest/tune/key-concepts.html
  For more information on Tune and Run configs, see
  https://docs.ray.io/en/latest/ray-air/tuner.html

  Args:
    function_name: BBOB function name.
    dimension: Dimension of BBOB function.
    shift: Shift of BBOB function.
    tune_config: Ray Tune Config.
    run_config: Ray Run Config.

  Returns:
  """
  experimenter_factory = experimenters.BBOBExperimenterFactory(
      name=function_name, dim=dimension
  )
  if shift is not None:
    experimenter_factory = experimenters.SingleObjectiveExperimenterFactory(
        base_factory=experimenter_factory, shift=shift
    )
  return run_tune_from_factory(experimenter_factory, tune_config, run_config)


def run_tune_from_factory(
    experimenter_factory: experimenters.ExperimenterFactory,
    tune_config: Optional[tune.TuneConfig] = None,
    run_config: Optional[air.RunConfig] = None,
) -> tune.result_grid.ResultGrid:
  """Runs Ray Tuners from an Experimenter Factory.

  See https://docs.ray.io/en/latest/tune/key-concepts.html
  For more information on Tune and Run configs, see
  https://docs.ray.io/en/latest/ray-air/tuner.html

  Args:
    experimenter_factory: Experimenter Factory.
    tune_config: Ray Tune Config.
    run_config: Ray Run Config.

  Returns:
  """
  experimenter = experimenter_factory()
  problem = experimenter.problem_statement()

  param_space = converters.SearchSpaceConverter.to_dict(problem.search_space)
  objective = converters.ExperimenterConverter.to_callable(experimenter)

  metric_info = problem.metric_information.item()
  if tune_config is None:
    tune_config = tune.TuneConfig()
  tune_config.metric = metric_info.name
  if metric_info.goal == pyvizier.ObjectiveMetricGoal.MINIMIZE:
    tune_config.mode = 'min'
  else:
    tune_config.mode = 'max'

  def objective_fn(config: pyvizier.ParameterDict) -> None:
    # Config contains parameter names to values and is autopopulated for each
    # Trial. Evaluation is static for BBOB so we simply loop.
    for _ in range(tune_config.num_samples):
      result_dict = objective(config)
      session.report(result_dict)

  tuner = tune.Tuner(
      objective_fn,
      param_space=param_space,
      run_config=run_config,
      tune_config=tune_config,
  )
  return tuner.fit()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for designer_optimizer."""

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.evolution import nsga2
from vizier._src.algorithms.optimizers import designer_optimizer
from vizier._src.algorithms.testing import optimizer_test_utils

from absl.testing import absltest


class DesignerOptimizerTest(absltest.TestCase):

  def test_smoke(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('a', 0, 1)

    designer_factory = quasi_random.QuasiRandomDesigner.from_problem
    optimizer_test_utils.assert_passes_on_random_single_metric_function(
        self,
        problem.search_space,
        designer_optimizer.DesignerAsOptimizer(designer_factory),
        np_random_seed=1)

  def test_bi_objective(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('a', 0, 1)

    designer_factory = nsga2.NSGA2Designer
    optimizer_test_utils.assert_passes_on_random_multi_metric_function(
        self,
        problem.search_space,
        designer_optimizer.DesignerAsOptimizer(designer_factory),
        np_random_seed=1,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for bbob."""

import sys

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class BbobTest(parameterized.TestCase):

  @parameterized.parameters(
      {'scale_type': vz.ScaleType.LINEAR}, {'scale_type': vz.ScaleType.LOG}
  )
  def test_bbob_problem_statement_scale_type(self, scale_type):
    problem = bbob.DefaultBBOBProblemStatement(10, scale_type=scale_type)
    for p in problem.search_space.parameters:
      self.assertEqual(p.scale_type, scale_type)

  def test_rotation(self):
    rotation = bbob._R(5, sys.maxsize, 'A', b'XAFDAF', 3)
    rotation2 = bbob._R(5, sys.maxsize, 'A', b'XAFDAF', 3)
    np.testing.assert_allclose(rotation.T @ rotation, np.eye(5), atol=1e-5)
    np.testing.assert_array_equal(rotation, rotation2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for hebo_gp_model."""

from absl import logging
import jax
from jax import config
from jax import numpy as jnp
import optax
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import hebo_gp_model
from vizier.jax import optimizers

from absl.testing import absltest


VizierHeboGaussianProcess = hebo_gp_model.VizierHeboGaussianProcess


class VizierHeboGaussianProcessTest(absltest.TestCase):

  def setUp(self):
    super(VizierHeboGaussianProcessTest, self).setUp()
    self.x_obs = jnp.array([[
        0.2941264, 0.29313548, 0.68817519, 0.37502566, 0.48356813, 0.34127283
    ], [
        0.66218224, 0.70770083, 0.6901334, 0.66787973, 0.5400858, 0.52721233
    ], [
        0.88469647, 0.50593371, 0.83160862, 0.58674892, 0.42145673, 0.31749428
    ], [
        0.39976682, 0.59517741, 0.73295106, 0.6084903, 0.54891015, 0.44338632
    ], [
        0.8354305, 0.87605574, 0.47855956, 0.48174861, 0.37685449, 0.38348768
    ], [
        0.55608455, 0.72781129, 0.52432913, 0.44291417, 0.3816395, 0.326599
    ], [
        0.24689187, 0.50979672, 0.67604857, 0.45172594, 0.34994392, 0.75239792
    ], [
        0.71007257, 0.60896354, 0.29270877, 0.74683367, 0.50169051, 0.74480515
    ], [
        0.9193235, 0.24393112, 0.63868591, 0.43271524, 0.43339578, 0.59413154
    ], [0.51850627, 0.62689204, 0.76134879, 0.65990021, 0.82350868, 0.7429215]],
                           dtype=jnp.float64)

    self.y_obs = jnp.array(
        [
            0.55552674,
            -0.29054829,
            -0.04703586,
            0.0217839,
            0.15445438,
            0.46654119,
            0.12255823,
            -0.19540335,
            -0.11772564,
            -0.44447326,
        ],
        dtype=jnp.float64,
    )[:, jnp.newaxis]

  def test_log_prob_and_loss(self):
    inputs = types.ModelInput(
        continuous=types.PaddedArray.as_padded(self.x_obs),
        categorical=types.PaddedArray.as_padded(
            jnp.zeros((self.x_obs.shape[0], 0)).astype(jnp.int32),
        ),
    )
    data = types.ModelData(
        features=inputs,
        labels=types.PaddedArray.as_padded(self.y_obs),
    )
    model = sp.CoroutineWithData(
        hebo_gp_model.VizierHeboGaussianProcess(), data=data
    )
    key, init_key, init_keys = jax.random.split(jax.random.PRNGKey(2), 3)
    init_params = model.setup(init_key)
    optimize = optimizers.OptaxTrain(optax.adam(5e-3), epochs=500, verbose=True)
    constraints = sp.get_constraints(model)
    params, metrics = optimize(
        init_params=jax.vmap(model.setup)(jax.random.split(init_keys, 20)),
        loss_fn=model.loss_with_aux,
        rng=key,
        constraints=constraints,
    )

    self.assertGreater(
        model.loss_with_aux(init_params), model.loss_with_aux(params)
    )
    losses_every_50 = metrics['loss'][::50]
    self.assertTrue((losses_every_50[1:] < losses_every_50[:-1]).all())

    logging.info('Optimal parameters: %s', params)
    final_loss = model.loss_with_aux(params)[0]
    logging.info('Optimal loss fn: %s', final_loss)
    self.assertLess(final_loss, 0.3)


if __name__ == '__main__':
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from absl.testing import parameterized
import equinox as eqx
import jax
from jax import numpy as jnp
import numpy as np
from vizier._src.jax import types as vt
from absl.testing import absltest


# pylint: disable=g-long-lambda
class PaddedArrayTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          factory=lambda arr: vt.PaddedArray.from_array(
              arr, arr.shape, fill_value=-1
          )
      ),
      dict(factory=vt.PaddedArray.as_padded),
  )
  def test_padding_to_the_same_shape(self, factory) -> None:
    """Tests the case where no padding dimensions are added.

    Args:
      factory: Takes an array and returns a padded array.
    """
    arr = jnp.ones([3, 2], dtype=jnp.int32)
    parr = factory(arr)
    # The padded shape matches the original shape, which means we can unpad
    # within a jit scope.
    self.assertSequenceEqual(
        eqx.filter_jit(parr.unpad)().shape,
        arr.shape,
    )

    np.testing.assert_array_equal(parr.padded_array, arr)
    np.testing.assert_array_equal(parr.is_missing[0], [False, False, False])
    np.testing.assert_array_equal(parr.is_missing[1], [False, False])
    np.testing.assert_array_equal(
        parr.replace_fill_value(-999).padded_array == -999, ~parr._mask
    )

  @parameterized.parameters(dict(fill_value=0), dict(fill_value=-1))
  def test_padding_to_a_different_shape(self, fill_value):
    arr = jnp.ones([3, 2], dtype=jnp.int32)
    # If the padded shape does not match the original shape, then we can't unpad
    # within a jit scope.
    target_shape = (5, 3)
    parr = vt.PaddedArray.from_array(arr, target_shape, fill_value=fill_value)
    with self.assertRaises(jax.errors.ConcretizationTypeError):
      eqx.filter_jit(parr.unpad)()

    expected = np.asarray(
        [
            [1, 1, 0],
            [1, 1, 0],
            [1, 1, 0],
            [0, 0, 0],
            [0, 0, 0],
        ],
        dtype=arr.dtype,
    )
    expected[expected == 0] = fill_value
    np.testing.assert_array_equal(parr.padded_array, expected)
    np.testing.assert_array_equal(parr.unpad(), arr)
    np.testing.assert_array_equal(
        parr.is_missing[0], [False, False, False, True, True]
    )
    np.testing.assert_array_equal(parr.is_missing[1], [False, False, True])
    np.testing.assert_array_equal(
        parr.replace_fill_value(-999).padded_array == -999, ~parr._mask
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Curve Analysis utilities for benchmarking and metalearning."""
# pylint: disable=unused-import

from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceComparator
from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceComparatorFactory
from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceCurve
from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceCurveConverter
from vizier._src.benchmarks.analyzers.convergence_curve import HypervolumeCurveConverter
from vizier._src.benchmarks.analyzers.convergence_curve import LogEfficiencyConvergenceCurveComparator
from vizier._src.benchmarks.analyzers.convergence_curve import LogEfficiencyConvergenceCurveComparatorFactory
from vizier._src.benchmarks.analyzers.convergence_curve import MultiMetricCurveConverter
from vizier._src.benchmarks.analyzers.convergence_curve import PercentageBetterConvergenceCurveComparator
from vizier._src.benchmarks.analyzers.convergence_curve import PercentageBetterConvergenceCurveComparatorFactory
from vizier._src.benchmarks.analyzers.convergence_curve import RestartingCurveConverter
from vizier._src.benchmarks.analyzers.convergence_curve import StatefulCurveConverter
from vizier._src.benchmarks.analyzers.plot_utils import plot_from_records
from vizier._src.benchmarks.analyzers.plot_utils import plot_mean_convergence
from vizier._src.benchmarks.analyzers.plot_utils import plot_median_convergence
from vizier._src.benchmarks.analyzers.state_analyzer import BenchmarkRecord
from vizier._src.benchmarks.analyzers.state_analyzer import BenchmarkRecordAnalyzer
from vizier._src.benchmarks.analyzers.state_analyzer import BenchmarkStateAnalyzer
from vizier._src.benchmarks.analyzers.state_analyzer import PlotElement



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for trial_regression_utils."""

import copy
from typing import Union

import lightgbm.sklearn as lightgbm
import numpy as np
from sklearn.model_selection import GridSearchCV
from vizier import algorithms as vza
from vizier import pyvizier
from vizier._src.algorithms.regression import trial_regression_utils
from vizier.pyvizier import converters

from absl.testing import absltest

_METRIC_NAME = 'objective_value'


def _create_trial_for_testing(
    learning_rate: float,
    steps: list[int],
    seconds: list[Union[int, float]],
    values: list[float],
    stop_reason: Union[None, str],
    trial_id: int,
    metric_name: str = _METRIC_NAME,
):
  measurements = []
  for i in range(len(steps)):
    measurements.append(
        pyvizier.Measurement(
            steps=steps[i],
            elapsed_secs=seconds[i],
            metrics={metric_name: values[i]},
        )
    )
  trial = pyvizier.Trial(
      id=trial_id,
      measurements=measurements,
      stopping_reason=stop_reason,
      parameters=pyvizier.ParameterDict({'learning_rate': learning_rate}),
      final_measurement=measurements[-1],
  )
  return trial


class TrialRegressionUtilsTest(absltest.TestCase):

  def test_preprocess_trial(self):
    steps = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    seconds = [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000]
    values = [0.1, 0.3, 0.4, 0.45, 0.48, 0.49, 0.5, 0.49, 0.51, 0.5]

    metric = pyvizier.MetricInformation(
        name=_METRIC_NAME, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
    )
    conv1 = converters.TimedLabelsExtractor(
        [
            converters.DefaultModelOutputConverter(
                metric, flip_sign_for_minimization_metrics=True
            ),
        ],
        timestamp='steps',
        value_extraction='raw',
    )
    trial = _create_trial_for_testing(
        0.3, steps, seconds, values, None, 1, _METRIC_NAME
    )
    conv2 = converters.TimedLabelsExtractor(
        [
            converters.DefaultModelOutputConverter(
                metric, flip_sign_for_minimization_metrics=True
            ),
        ],
        timestamp='elapsed_secs',
        value_extraction='raw',
    )

    conv_trial1 = trial_regression_utils.TrialData.from_trial(
        trial=trial,
        learning_rate_param_name='learning_rate',
        metric_name=_METRIC_NAME,
        converter=conv1,
    )
    conv_trial2 = trial_regression_utils.TrialData.from_trial(
        trial=trial,
        learning_rate_param_name='learning_rate',
        metric_name=_METRIC_NAME,
        converter=conv2,
    )
    self.assertListEqual(steps, conv_trial1.steps)
    self.assertSequenceAlmostEqual(values, conv_trial1.objective_values)
    self.assertListEqual(seconds, conv_trial2.steps)
    self.assertSequenceAlmostEqual(values, conv_trial2.objective_values)

  def test_sort_dedupe_measurements(self):
    # Case 1: no duplicates.
    steps = [10, 20]
    values = [0.1, 0.2]
    actual_steps, actual_values = (
        trial_regression_utils._sort_dedupe_measurements(steps, values)
    )
    self.assertListEqual(steps, actual_steps)
    self.assertListEqual(values, actual_values)

    # Case 2: with duplicates, and measurement order needs correction.
    steps = [10, 20, 10]
    values = [0.1, 0.2, 0.2]
    expected_steps = [10, 20]
    expected_values = [0.2, 0.2]
    actual_steps, actual_values = (
        trial_regression_utils._sort_dedupe_measurements(steps, values)
    )
    self.assertListEqual(expected_steps, actual_steps)
    self.assertListEqual(expected_values, actual_values)

  def test_generate_interpolation_fn_from_trial(self):
    steps = [10, 20]
    values = [0.1, 0.5]
    fn = trial_regression_utils._generate_interpolation_fn_from_trial(
        steps, values
    )
    self.assertEqual(fn(steps[0]), values[0])
    self.assertEqual(fn(steps[1]), values[1])
    self.assertGreater(fn(15), values[0])
    self.assertLess(fn(15), values[1])

  def test_extrapolate_trial(self):
    steps = [10, 20, 30]
    values = [0.1, 0.3, 0.5]

    expected_steps = [10, 20, 30, 100]
    expected_values = [0.1, 0.3, 0.5, 0.5]

    trial = trial_regression_utils.TrialData(
        id=1,
        learning_rate=0.1,
        final_objective=values[-1],
        steps=steps,
        objective_values=values,
    )

    expected_trial = copy.deepcopy(trial)

    # Case 1: enough steps, no extension needed.
    trial.extrapolate_trial_objective_value(30)
    self.assertEqual(trial, expected_trial)

    # Case 2: fewer steps, extension needed.
    expected_trial.extrapolate_trial_objective_value(100)
    self.assertEqual(expected_trial.id, trial.id)
    self.assertEqual(expected_trial.learning_rate, trial.learning_rate)
    self.assertEqual(expected_trial.final_objective, trial.final_objective)
    self.assertListEqual(expected_steps, expected_trial.steps)
    self.assertListEqual(expected_values, expected_trial.objective_values)


class GBMAutoRegressorTest(absltest.TestCase):

  def test_create_features_from_trial(self):
    steps = [10, 20, 30]
    values = [0.1, 0.3, 0.5]

    trial = trial_regression_utils.TrialData(
        id=1,
        learning_rate=0.1,
        final_objective=values[-1],
        steps=steps,
        objective_values=values,
    )
    metric = pyvizier.MetricInformation(
        name=_METRIC_NAME, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
    )
    conv = converters.TimedLabelsExtractor(
        [
            converters.DefaultModelOutputConverter(
                metric, flip_sign_for_minimization_metrics=True
            ),
        ],
        timestamp='steps',
        value_extraction='raw',
    )

    gbm = trial_regression_utils.GBMAutoRegressor(
        min_points=3,
        target_step=40,
        learning_rate_param_name='learning_rate',
        metric_name=_METRIC_NAME,
        converter=conv,
    )
    features = gbm._create_features_from_trial(trial, end_index=2)
    expected_features = [0.1, 10, 0.5, 20, 0.3, 30, 0.1]
    self.assertListEqual(features, expected_features)

  def test_gbmautoregressor_train_predict(self):
    metric = pyvizier.MetricInformation(
        name=_METRIC_NAME, goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
    )
    conv = converters.TimedLabelsExtractor(
        [
            converters.DefaultModelOutputConverter(
                metric, flip_sign_for_minimization_metrics=True
            ),
        ],
        timestamp='steps',
        value_extraction='raw',
    )
    gbm = trial_regression_utils.GBMAutoRegressor(
        target_step=50,
        min_points=2,
        learning_rate_param_name='learning_rate',
        metric_name=_METRIC_NAME,
        converter=conv,
        random_state=111,
    )
    with self.subTest('TestModelCreation'):
      steps = [10, 20, 30, 40, 50]
      values1 = [0.1, 0.3, 0.5, 0.6, 0.66]
      values2 = [0.1, 0.25, 0.4, 0.5, 0.55]
      values3 = [0.1, 0.2, 0.35, 0.48, 0.6]

      trial1 = trial_regression_utils.TrialData(
          id=1,
          learning_rate=0.1,
          final_objective=values1[-1],
          steps=steps,
          objective_values=values1,
      )
      pytrial1 = _create_trial_for_testing(
          learning_rate=0.1,
          steps=steps,
          seconds=steps,
          values=values1,
          trial_id=1,
          stop_reason=None,
          metric_name=_METRIC_NAME,
      )
      trial2 = trial_regression_utils.TrialData(
          id=2,
          learning_rate=0.3,
          final_objective=values2[-1],
          steps=steps,
          objective_values=values2,
      )
      pytrial2 = _create_trial_for_testing(
          learning_rate=0.3,
          steps=steps,
          seconds=steps,
          values=values2,
          trial_id=2,
          stop_reason=None,
          metric_name=_METRIC_NAME,
      )
      trial3 = trial_regression_utils.TrialData(
          id=3,
          learning_rate=0.2,
          final_objective=values3[-1],
          steps=steps,
          objective_values=values3,
      )
      pytrial3 = _create_trial_for_testing(
          learning_rate=0.2,
          steps=steps,
          seconds=steps,
          values=values3,
          trial_id=3,
          stop_reason=None,
          metric_name=_METRIC_NAME,
      )
      feat_mat = []
      targets = []
      for trial in [trial1, trial2, trial3]:
        for i, step in enumerate(trial.steps):
          if i < 1 or step >= 50:
            continue
          features = gbm._create_features_from_trial(trial, 2)
          feat_mat.append(features)
          targets.append(trial.objective_values[-1])
      feat_mat = np.array(feat_mat)
      targets = np.array(targets)
      gbdt_param_grid = {'max_depth': [2, 3], 'n_estimators': [50, 100]}
      gbdt_cv = GridSearchCV(
          lightgbm.LGBMRegressor(random_state=111), gbdt_param_grid, cv=2
      )
      gbdt_cv = gbdt_cv.fit(feat_mat, targets)
      best_params = gbdt_cv.best_params_
      ideal_model = lightgbm.LGBMRegressor(**best_params, random_state=111)
      ideal_model = ideal_model.fit(feat_mat, targets)
      gbm.train(vza.CompletedTrials(trials=[pytrial1, pytrial2, pytrial3]))
      pred = gbm._model.predict(feat_mat[[1], :])
      pred_expected = ideal_model.predict(feat_mat[[1], :])
      self.assertAlmostEqual(pred, pred_expected)

    with self.subTest('TestPrediction'):
      steps = [10, 20, 30, 40]
      values = [0.1, 0.3, 0.5, 0.6]
      trial_pred = trial_regression_utils.TrialData(
          id=2,
          learning_rate=0.1,
          final_objective=values[-1],
          steps=steps,
          objective_values=values,
      )
      pytrial_pred = _create_trial_for_testing(
          learning_rate=0.1,
          steps=steps,
          seconds=steps,
          values=values,
          trial_id=2,
          stop_reason=None,
          metric_name=_METRIC_NAME,
      )
      pred = gbm.predict(pytrial_pred)
      features = gbm._create_features_from_trial(trial_pred, end_index=2)
      features = np.array(features).reshape(1, -1)
      pred_expected = ideal_model.predict(features)[0]
      self.assertAlmostEqual(pred, pred_expected)

  def test_converter_in_trial_hallucinator(self):
    steps = [10, 20, 30, 40, 50]
    values = [0.1, 0.3, 0.5, 0.6, 0.66]
    pytrial = _create_trial_for_testing(
        learning_rate=0.1,
        steps=steps,
        seconds=steps,
        values=values,
        trial_id=1,
        stop_reason=None,
        metric_name=_METRIC_NAME,
    )
    options = trial_regression_utils.HallucinationOptions(max_steps=50)
    problem = pyvizier.ProblemStatement()
    problem.search_space.root.add_float_param(
        'learning_rate', min_value=0.0, max_value=1.0, default_value=0.5
    )
    problem.search_space.root.add_int_param(
        'num_layers', min_value=1, max_value=5
    )
    problem.metric_information = [
        pyvizier.MetricInformation(
            name=_METRIC_NAME, goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
        )
    ]
    hallucinator = trial_regression_utils.GBMTrialHallucinator(
        study_config=problem, options=options
    )
    trial = trial_regression_utils.TrialData.from_trial(
        trial=pytrial,
        learning_rate_param_name='learning_rate',
        metric_name=_METRIC_NAME,
        converter=hallucinator._converter,
    )
    self.assertSequenceAlmostEqual(
        trial.objective_values, [0.1, 0.3, 0.5, 0.6, 0.66]
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from absl.testing import parameterized
import numpy as np
from vizier._src.algorithms.designers import grid
from vizier._src.benchmarks.experimenters.synthetic import simplekd
from vizier._src.benchmarks.runners import benchmark_runner
from vizier._src.benchmarks.runners import benchmark_state
from absl.testing import absltest


class SimpleKDTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(best_category='corner'),
      dict(best_category='center'),
      dict(best_category='mixed'),
  )
  def test_sweep(self, best_category: simplekd.SimpleKDCategory) -> None:
    experimenter = simplekd.SimpleKD(best_category)
    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[
            benchmark_runner.GenerateSuggestions(),
            benchmark_runner.EvaluateActiveTrials(),
        ],
        num_repeats=300,
    )
    state = benchmark_state.BenchmarkState(
        experimenter,
        benchmark_state.PolicySuggester.from_designer_factory(
            experimenter.problem_statement(),
            grid.GridSearchDesigner.from_problem,
        ),
    )
    runner.run(state)

  @parameterized.parameters(
      dict(best_category='corner'),
      dict(best_category='center'),
      dict(best_category='mixed'),
  )
  def test_compute_optimal_objective(
      self, best_category: simplekd.SimpleKDCategory
  ) -> None:
    exptr_simple4d = simplekd.SimpleKD(best_category)
    categorical_values = ['corner', 'center', 'mixed']
    discrete_values = (1, 2, 5, 6, 8)
    integer_values = [1, 2, 3]
    continuous_values = np.linspace(-1, 1, 1001)

    opt_value = np.nan
    for cat_value in categorical_values:
      for disc_value in discrete_values:
        for cont_value in continuous_values:
          for int_value in integer_values:
            values = {
                'categorical': [cat_value],
                'discrete': [disc_value],
                'int': [int_value],
                'float': [cont_value],
            }
            opt_value = np.nanmax([exptr_simple4d._compute(values), opt_value])

    self.assertAlmostEqual(opt_value, exptr_simple4d.optimal_objective)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for json_utils."""

import json
import numpy as np

from vizier.utils import json_utils
from absl.testing import absltest
from absl.testing import parameterized


class JsonUtilsTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(shape=[3, 0]),
      dict(shape=[3, 5]),
  )
  def test_dump_and_recover(self, shape):
    original = {'a': np.zeros(shape), 'b': 3}
    dumped = json.dumps(original, cls=json_utils.NumpyEncoder)
    loaded = json.loads(dumped, cls=json_utils.NumpyDecoder)
    np.testing.assert_array_equal(original['a'], loaded['a'])
    self.assertEqual(original['b'], loaded['b'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Assertions for testing experimenters."""

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.experimenters import experimenter as experimenter_lib
from vizier._src.benchmarks.runners import benchmark_runner
from vizier._src.benchmarks.runners import benchmark_state


def assert_evaluates_random_suggestions(
    test,
    experimenter: experimenter_lib.Experimenter,
) -> None:
  """Asserts that random suggestions from the search space are valid."""
  runner = benchmark_runner.BenchmarkRunner(
      [benchmark_runner.GenerateAndEvaluate(10)], num_repeats=1
  )

  state = benchmark_state.BenchmarkState(
      experimenter=experimenter,
      algorithm=benchmark_state.PolicySuggester.from_designer_factory(
          experimenter.problem_statement(), random.RandomDesigner.from_problem
      ),
  )

  runner.run(state)

  test.assertLen(
      state.algorithm.supporter.GetTrials(
          status_matches=vz.TrialStatus.COMPLETED
      ),
      10,
  )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for attrs_utils."""

from typing import Any

import attrs
import numpy as np

from vizier.utils import attrs_utils
from absl.testing import absltest
from absl.testing import parameterized


class ValidatorsTest(parameterized.TestCase):

  @parameterized.parameters(
      (attrs_utils.assert_not_empty, [], False),
      (attrs_utils.assert_not_empty, [1], True),
      (attrs_utils.assert_not_negative, 0, -1),
      (attrs_utils.assert_not_negative, 0, True),
      (attrs_utils.assert_not_negative, 1, True),
      (attrs_utils.assert_not_none, None, False),
      (attrs_utils.assert_not_none, 0, True),
  )
  def test_validator(self, validator, value: Any, result: bool):
    self.assertValidatorWorksAsIntended(validator, value, result)

  def assertValidatorWorksAsIntended(self, validator, value: Any, result: bool):

    @attrs.define
    class Test:
      x = attrs.field(validator=validator)

    if result:
      Test(value)
    else:
      with self.assertRaises(ValueError):
        Test(value)

  @parameterized.parameters(
      (0.0, 1.0, 0.5, True),
      (0.0, 1.0, 1.0, True),
      (0.0, 1.0, 0.0, True),
      (0.0, 1.0, 5.0, False),
  )
  def test_between_validator(
      self, low: float, high: float, value: float, result: bool
  ):
    self.assertValidatorWorksAsIntended(
        attrs_utils.assert_between(low, high), value, result
    )

  @parameterized.parameters(
      (r'[^\/]+', 'good', True),
      (r'[^\/]+', 'bad/', False),
      (r'[^\/]+', '', False),
  )
  def test_regex_validator(self, regex: str, value: str, result: bool):
    self.assertValidatorWorksAsIntended(
        attrs_utils.assert_re_fullmatch(regex), value, result)

  def test_good_shape_none(self):
    _ShapeEqualsTestAttr(np.zeros([3, 5]), None)
    _ShapeEqualsTestAttr(np.zeros([3, 0]), None)

  def test_bad_shape(self):
    with self.assertRaises(ValueError):
      _ShapeEqualsTestAttr(np.zeros([3, 2]), 4)


@attrs.define
class _ShapeEqualsTestAttr:
  x = attrs.field(validator=attrs_utils.shape_equals(lambda v: (3, v.d)))
  d = attrs.field()


class ShapeEqualsTest(absltest.TestCase):

  def test_good_shape(self):
    _ShapeEqualsTestAttr(np.zeros([3, 2]), 2)

  def test_good_shape_none(self):
    _ShapeEqualsTestAttr(np.zeros([3, 5]), None)
    _ShapeEqualsTestAttr(np.zeros([3, 0]), None)

  def test_bad_shape(self):
    with self.assertRaises(ValueError):
      _ShapeEqualsTestAttr(np.zeros([3, 2]), 4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Core algorithm modules."""

from vizier._src.algorithms.core.abstractions import ActiveTrials
from vizier._src.algorithms.core.abstractions import CompletedTrials
from vizier._src.algorithms.core.abstractions import Designer
from vizier._src.algorithms.core.abstractions import DesignerFactory
from vizier._src.algorithms.core.abstractions import PartiallySerializableDesigner
from vizier._src.algorithms.core.abstractions import Prediction
from vizier._src.algorithms.core.abstractions import Predictor
from vizier._src.algorithms.core.abstractions import SerializableDesigner
from vizier._src.algorithms.optimizers.base import BatchTrialScoreFunction
from vizier._src.algorithms.optimizers.base import BranchSelection
from vizier._src.algorithms.optimizers.base import BranchSelector
from vizier._src.algorithms.optimizers.base import BranchThenOptimizer
from vizier._src.algorithms.optimizers.base import GradientFreeOptimizer
from vizier._src.algorithms.policies.designer_policy import DesignerPolicy
from vizier._src.algorithms.policies.designer_policy import InRamDesignerPolicy
from vizier._src.algorithms.policies.designer_policy import PartiallySerializableDesignerPolicy
from vizier._src.algorithms.policies.designer_policy import SerializableDesignerPolicy



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vectorized_base."""

from typing import Optional

import chex
import jax
from jax import numpy as jnp
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import types
from vizier.pyvizier import converters

from absl.testing import absltest
from absl.testing import parameterized

# pylint: disable=g-long-lambda


@chex.dataclass(frozen=True)
class FakeIncrementVectorizedStrategyState:
  """State for FakeIncrementVectorizedStrategy."""

  iterations: int


class FakeIncrementVectorizedStrategy(
    vb.VectorizedStrategy[FakeIncrementVectorizedStrategyState]
):
  """Fake vectorized strategy with incrementing suggestions."""

  def __init__(self, *args, **kwargs):
    pass

  def suggest(
      self,
      seed: jax.Array,
      state: FakeIncrementVectorizedStrategyState,
      n_parallel: int = 1,
  ) -> vb.VectorizedOptimizerInput:
    # The following structure allows to test the top K results.
    i = state.iterations
    suggestions = (
        jnp.array([
            [i % 10, i % 10],
            [(i + 1) % 10, (i + 1) % 10],
            [(i + 2) % 10, (i + 2) % 10],
            [(i + 3) % 10, (i + 3) % 10],
            [(i + 4) % 10, (i + 4) % 10],
        ])
        / 10
    )
    return vb.VectorizedOptimizerInput(
        jnp.repeat(suggestions[:, jnp.newaxis, :], n_parallel, axis=1),
        jnp.zeros([5, n_parallel, 0], dtype=types.INT_DTYPE),
    )

  @property
  def suggestion_batch_size(self) -> int:
    return 5

  def update(
      self,
      seed: jax.Array,
      state: FakeIncrementVectorizedStrategyState,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: types.Array,
  ) -> FakeIncrementVectorizedStrategyState:
    return FakeIncrementVectorizedStrategyState(iterations=state.iterations + 5)

  def init_state(
      self,
      seed: jax.Array,
      n_parallel: int = 1,
      *,
      prior_features: Optional[vb.VectorizedOptimizerInput] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> FakeIncrementVectorizedStrategyState:
    del seed
    return FakeIncrementVectorizedStrategyState(iterations=0)


# pylint: disable=unused-argument
def fake_increment_strategy_factory(
    converter: converters.TrialToModelInputConverter,
    suggestion_batch_size: int,
) -> vb.VectorizedStrategy:
  return FakeIncrementVectorizedStrategy()


@chex.dataclass(frozen=True)
class FakePriorTrialsStrategyState:
  """State for FakeIncrementVectorizedStrategy."""

  features: vb.VectorizedOptimizerInput
  rewards: types.Array


class FakePriorTrialsVectorizedStrategy(
    vb.VectorizedStrategy[FakePriorTrialsStrategyState]
):
  """Fake vectorized strategy to test prior trials."""

  def init_state(
      self,
      seed: jax.Array,
      n_parallel: int = 1,
      *,
      prior_features: Optional[vb.VectorizedOptimizerInput] = None,
      prior_rewards: Optional[types.Array] = None,
  ):
    if prior_rewards is not None and len(prior_rewards.shape) != 1:
      raise ValueError('Expected seed labels to have 1D dimension!')
    return FakePriorTrialsStrategyState(
        features=prior_features, rewards=prior_rewards
    )

  def suggest(
      self,
      seed: jax.Array,
      state: FakePriorTrialsStrategyState,
      n_parallel: int = 1,
  ) -> vb.VectorizedOptimizerInput:
    return vb.VectorizedOptimizerInput(
        continuous=state.features.continuous[
            jnp.argmax(state.rewards, axis=-1)
        ][jnp.newaxis, :],
        categorical=jnp.zeros([1, n_parallel, 0], dtype=types.INT_DTYPE),
    )

  @property
  def suggestion_batch_size(self) -> int:
    return 1

  def update(
      self,
      seed: jax.Array,
      state: FakePriorTrialsStrategyState,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: types.Array,
  ) -> FakePriorTrialsStrategyState:
    return state


# pylint: disable=unused-argument
def fake_prior_trials_strategy_factory(
    converter: converters.TrialToModelInputConverter,
    suggestion_batch_size: int,
) -> vb.VectorizedStrategy:
  return FakePriorTrialsVectorizedStrategy()


class VectorizedBaseTest(parameterized.TestCase):

  @parameterized.parameters(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
  def test_optimize_candidates_len(self, count):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    score_fn = lambda x, _: jnp.sum(x.continuous.padded_array, axis=-1)
    optimizer = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_increment_strategy_factory,
        max_evaluations=100,
    )(converter=converter)
    res_array = optimizer(score_fn=score_fn, count=count)
    res = vb.best_candidates_to_trials(res_array, converter=converter)
    self.assertLen(res, count)

  @parameterized.parameters(
      (1, 3),
      (2, 4),
  )
  def test_optimize_parallel_candidates_len(self, count, n_parallel):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    score_fn = lambda x, _: jnp.sum(x.continuous.padded_array, axis=(-1, -2))
    optimizer = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_increment_strategy_factory,
        max_evaluations=100,
    )(converter=converter)
    res_array = optimizer(score_fn=score_fn, count=count, n_parallel=n_parallel)
    res = vb.best_candidates_to_trials(res_array, converter=converter)
    self.assertLen(res, count * n_parallel)

  @parameterized.parameters(True, False)
  def test_best_candidates_count_is_1(self, use_fori):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    score_fn = lambda x, _: -jnp.max(
        jnp.square(x.continuous.padded_array - 0.52), axis=-1
    )
    strategy_factory = FakeIncrementVectorizedStrategy
    optimizer = vb.VectorizedOptimizerFactory(
        strategy_factory=strategy_factory,
        suggestion_batch_size=5,
        max_evaluations=10,
        use_fori=use_fori,
    )(converter=converter)
    best_candidates_array = optimizer(score_fn=score_fn, count=1)
    best_candidates = vb.best_candidates_to_trials(
        best_candidates_array, converter=converter
    )
    # check the best candidate
    self.assertEqual(best_candidates[0].parameters['f1'].value, 0.5)
    self.assertEqual(best_candidates[0].parameters['f2'].value, 0.5)
    self.assertAlmostEqual(
        best_candidates[0].final_measurement.metrics['acquisition'].value,
        -((0.5 - 0.52) ** 2),
    )

  @parameterized.parameters(True, False)
  def test_best_candidates_count_is_3(self, use_fori):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    score_fn = lambda x, _: -jnp.max(
        jnp.square(x.continuous.padded_array - 0.52), axis=-1
    )
    optimizer = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_increment_strategy_factory,
        suggestion_batch_size=5,
        max_evaluations=10,
        use_fori=use_fori,
    )(converter=converter)
    best_candidates_array = optimizer(score_fn=score_fn, count=3)
    best_candidates = vb.best_candidates_to_trials(
        best_candidates_array, converter=converter
    )
    # check 1st best candidate
    self.assertAlmostEqual(best_candidates[0].parameters['f1'].value, 0.5)
    self.assertAlmostEqual(best_candidates[0].parameters['f2'].value, 0.5)
    self.assertAlmostEqual(
        best_candidates[0].final_measurement.metrics['acquisition'].value,
        -((0.5 - 0.52) ** 2),
    )
    # check 2nd best candidate
    self.assertAlmostEqual(best_candidates[1].parameters['f1'].value, 0.6)
    self.assertAlmostEqual(best_candidates[1].parameters['f2'].value, 0.6)
    self.assertAlmostEqual(
        best_candidates[1].final_measurement.metrics['acquisition'].value,
        -((0.6 - 0.52) ** 2),
    )
    # check 3rd best candidate
    self.assertAlmostEqual(best_candidates[2].parameters['f1'].value, 0.4)
    self.assertAlmostEqual(best_candidates[2].parameters['f2'].value, 0.4)
    self.assertAlmostEqual(
        best_candidates[2].final_measurement.metrics['acquisition'].value,
        -((0.4 - 0.52) ** 2),
    )

  def test_vectorized_optimizer_factory(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_increment_strategy_factory,
        suggestion_batch_size=5,
        max_evaluations=1000,
    )
    optimizer = optimizer_factory(converter)
    self.assertEqual(optimizer.max_evaluations, 1000)
    self.assertEqual(optimizer.suggestion_batch_size, 5)

  @parameterized.parameters(True, False)
  def test_prior_trials(self, use_fori):
    """Test that the optimizer can correctly parsae and pass seed trials."""
    optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_prior_trials_strategy_factory,
        suggestion_batch_size=5,
        max_evaluations=100,
        use_fori=use_fori,
    )

    study_config = vz.ProblemStatement(
        metric_information=[
            vz.MetricInformation(
                name='obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = study_config.search_space.root
    root.add_float_param('x1', 0.0, 10.0)
    root.add_float_param('x2', 0.0, 10.0)
    converter = converters.TrialToModelInputConverter.from_problem(study_config)
    optimizer = optimizer_factory(converter)

    trial1 = vz.Trial(parameters={'x1': 1, 'x2': 1})
    measurement1 = vz.Measurement(metrics={'obj': vz.Metric(value=-10.33)})
    trial1.complete(measurement1, inplace=True)

    trial2 = vz.Trial(parameters={'x1': 2, 'x2': 2})
    measurement2 = vz.Measurement(metrics={'obj': vz.Metric(value=5.0)})
    trial2.complete(measurement2, inplace=True)

    prior_features = vb.trials_to_sorted_array(
        [trial1, trial2, trial1], converter=converter
    )
    best_trial_array = optimizer(
        lambda x, _: -jnp.max(
            jnp.square(x.continuous.padded_array - 0.52), axis=-1
        ),
        count=1,
        prior_features=prior_features,
    )
    best_trial = vb.best_candidates_to_trials(
        best_trial_array, converter=converter
    )
    self.assertEqual(best_trial[0].parameters['x1'].value, 2)
    self.assertEqual(best_trial[0].parameters['x2'].value, 2)

    best_trial_array = optimizer(
        lambda x, _: -jnp.max(
            jnp.square(x.continuous.padded_array - 0.52), axis=-1
        ),
        count=1,
        prior_features=vb.trials_to_sorted_array([trial1], converter=converter),
    )
    best_trial = vb.best_candidates_to_trials(
        best_trial_array, converter=converter
    )
    self.assertEqual(best_trial[0].parameters['x1'].value, 1)
    self.assertEqual(best_trial[0].parameters['x2'].value, 1)

  @parameterized.parameters(True, False)
  def test_prior_trials_parallel(self, use_fori):
    optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_prior_trials_strategy_factory,
        suggestion_batch_size=5,
        max_evaluations=100,
    )

    study_config = vz.ProblemStatement(
        metric_information=[
            vz.MetricInformation(
                name='obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = study_config.search_space.root
    root.add_float_param('x1', 0.0, 10.0)
    root.add_float_param('x2', 0.0, 10.0)
    root.add_float_param('x3', 0.0, 10.0)
    converter = converters.TrialToModelInputConverter.from_problem(study_config)
    optimizer = optimizer_factory(converter)
    prior_features = types.ModelInput(
        continuous=types.PaddedArray.as_padded(
            jax.random.uniform(jax.random.PRNGKey(0), (14, 3))
        ),
        categorical=types.PaddedArray.as_padded(
            jnp.zeros([14, 0], dtype=types.INT_DTYPE)
        ),
    )
    suggestions = optimizer(
        lambda x, _: -jnp.max(
            jnp.square(x.continuous.padded_array - 0.52), axis=(-1, -2)
        ),
        prior_features=prior_features,
        n_parallel=2,
    )
    self.assertSequenceEqual(suggestions.features.continuous.shape, (1, 2, 3))
    self.assertSequenceEqual(suggestions.rewards.shape, (1,))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for nsga2."""

import datetime
from typing import Optional

from absl import logging

import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz

from vizier._src.algorithms.evolution import nsga2
from vizier._src.algorithms.evolution import templates
from vizier.testing import test_studies

from absl.testing import absltest

np.set_printoptions(precision=3)


def nsga2_on_all_types(
    population_size: int = 50, eviction_limit: Optional[int] = None
) -> templates.CanonicalEvolutionDesigner[nsga2.Population, nsga2.Offspring]:
  problem = vz.ProblemStatement(
      search_space=test_studies.flat_space_with_all_types()
  )
  problem.metric_information.extend([
      vz.MetricInformation(name='m1', goal=vz.ObjectiveMetricGoal.MAXIMIZE),
      vz.MetricInformation(name='m2', goal=vz.ObjectiveMetricGoal.MINIMIZE),
      vz.MetricInformation(
          name='s1',
          goal=vz.ObjectiveMetricGoal.MAXIMIZE,
          safety_threshold=1.0,
      ),
      vz.MetricInformation(
          name='s2',
          goal=vz.ObjectiveMetricGoal.MINIMIZE,
          safety_threshold=1.0,
      ),
  ])

  algorithm = nsga2.NSGA2Designer(
      problem,
      population_size,
      first_survival_after=population_size,
      eviction_limit=eviction_limit,
      seed=0,
  )
  return algorithm


class Nsga2Test(absltest.TestCase):

  def test_survival_by_pareto_rank(self):
    algorithm = nsga2_on_all_types(3)
    # Trial 0 is the only point on the frontier.
    trial0 = vz.Trial(id=0)
    trial0.complete(
        vz.Measurement({'m1': 1.0, 'm2': 0.0, 's1': 2.0, 's2': 0.0})
    )

    # 4 safe trials with the same pareto rank. Crowding distance is computed
    # among them to break ties. Trial 3 is less "crowded" than Trial 2.
    trial1 = vz.Trial(id=1)
    trial1.complete(
        vz.Measurement({'m1': 0.0, 'm2': -1.0, 's1': 2.0, 's2': 0.0})
    )
    trial2 = vz.Trial(id=2)
    trial2.complete(
        vz.Measurement({'m1': 0.5, 'm2': -0.5, 's1': 2.0, 's2': 0.0})
    )
    trial3 = vz.Trial(id=3)
    trial3.complete(
        vz.Measurement({'m1': 0.2, 'm2': -0.2, 's1': 2.0, 's2': 0.0})
    )
    trial4 = vz.Trial(id=4)
    trial4.complete(
        vz.Measurement({'m1': 0.3, 'm2': -0.3, 's1': 2.0, 's2': 0.0})
    )

    trials = vza.CompletedTrials([trial0, trial1, trial2, trial3, trial4])
    algorithm.update(trials, vza.ActiveTrials())
    self.assertSetEqual(set(algorithm.population.trial_ids), {0, 1, 2})

  def test_survival_by_crowding_distance(self):
    algorithm = nsga2_on_all_types(4)
    # Trial 0 is the only point on the frontier.
    trial0 = vz.Trial(id=0)
    trial0.complete(
        vz.Measurement({'m1': 1.001, 'm2': -1.001, 's1': 2.0, 's2': 0.0})
    )

    # 4 safe trials with the same pareto rank. Crowding distance is computed
    # among them to break ties. Trial 3 is less "crowded" than Trial 2.
    trial1 = vz.Trial(id=1)
    trial1.complete(
        vz.Measurement({'m1': 1.0, 'm2': 0.0, 's1': 2.0, 's2': 0.9})
    )
    trial2 = vz.Trial(id=2)
    trial2.complete(
        vz.Measurement({'m1': 0.9, 'm2': -0.1, 's1': 2.0, 's2': 0.9})
    )
    trial3 = vz.Trial(id=3)
    trial3.complete(
        vz.Measurement({'m1': 0.5, 'm2': -0.5, 's1': 2.0, 's2': 0.9})
    )
    trial4 = vz.Trial(id=4)
    trial4.complete(
        vz.Measurement({'m1': 0.0, 'm2': -1.0, 's1': 2.0, 's2': 0.9})
    )

    trials = vza.CompletedTrials([trial0, trial1, trial2, trial3, trial4])
    algorithm.update(trials, vza.ActiveTrials())
    self.assertSetEqual(set(algorithm.population.trial_ids), {0, 1, 3, 4})

  def test_survival_by_safety(self):
    algorithm = nsga2_on_all_types(3)

    # Trial 1 violates 's1'.
    trial1 = vz.Trial(id=1)
    trial1.complete(
        vz.Measurement({'m1': 1.0, 'm2': 1.0, 's1': 0.0, 's2': 0.0})
    )
    # Trial 2 violates no constraints.
    trial2 = vz.Trial(id=2)
    trial2.complete(
        vz.Measurement({'m1': 0.9, 'm2': 0.9, 's1': 2.0, 's2': 0.0})
    )
    # Trial 3 and 4 violate both 's1' and 's2'.
    trial3 = vz.Trial(id=3)
    trial3.complete(
        vz.Measurement({'m1': 0.5, 'm2': 0.5, 's1': 0.0, 's2': 2.0})
    )
    trial4 = vz.Trial(id=4)
    trial4.complete(
        vz.Measurement({'m1': 0.0, 'm2': 0.0, 's1': 0.0, 's2': 2.0})
    )

    trials = vza.CompletedTrials([trial1, trial2, trial3, trial4])
    algorithm.update(trials, vza.ActiveTrials())
    self.assertSetEqual(set(algorithm.population.trial_ids), {1, 2, 4})

  def test_comprehensive_sanity_check(self):
    algorithm = nsga2_on_all_types(5, eviction_limit=3)

    tid = 1
    for i in range(10):
      if i == 8:
        dumped = algorithm.dump()
      tick = datetime.datetime.now()
      suggestions = algorithm.suggest()
      tock = datetime.datetime.now()
      logging.info('Iteration %s: Suggestion took %s.', i, tock - tick)
      trials = []
      for t in suggestions:
        trials.append(
            t.to_trial(tid).complete(
                vz.Measurement(
                    metrics={
                        'm1': np.random.random(),
                        'm2': np.random.random(),
                        's1': np.random.uniform(1.0, 2.0),
                        's2': np.random.uniform(
                            0.0, 1.0
                        ),  # fails with 25% probability
                    }
                )
            )
        )
        tid += 1
      tick = datetime.datetime.now()
      logging.info(
          'Suggesitons evaluated: %s', '\n'.join(repr(t) for t in trials)
      )
      algorithm.update(vza.CompletedTrials(trials), vza.ActiveTrials())
      tock = datetime.datetime.now()
      logging.info(
          (
              'Iteration %s: Update took %s.\nPopulation(in array'
              ' format):%s\nAges:%s'
          ),
          i,
          tock - tick,
          algorithm.population.xs,
          algorithm.population.ages,
      )
      tick = tock

    self.assertTrue(np.all(algorithm.population.ages <= 3))

    ys = algorithm.population.ys
    pareto = algorithm.population[nsga2._pareto_rank(ys) == 0]
    logging.info('Pareto frontier %s %s', pareto.xs, pareto.ys)

    # Smoke test dump-load.
    algorithm.load(dumped)

  def test_seeding(self):
    algorithm_1 = nsga2_on_all_types(10)
    algorithm_2 = nsga2_on_all_types(10)

    self.assertEqual(algorithm_1.suggest(10), algorithm_2.suggest(10))

    completed_trials = []

    for tid in range(1, 11):
      trial = vz.Trial(id=tid)

      trial.complete(
          vz.Measurement({
              'm1': np.random.random(),
              'm2': np.random.random(),
              's1': np.random.random(),
              's2': np.random.random(),
          })
      )

      completed_trials.append(trial)

    trials = vza.CompletedTrials(completed_trials)
    algorithm_1.update(trials, vza.ActiveTrials())
    algorithm_2.update(trials, vza.ActiveTrials())

    self.assertEqual(algorithm_1.suggest(10), algorithm_2.suggest(10))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import discretizing_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class DiscretizingExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere), ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated), ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar), ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass), ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel), ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek), ('Gallagher101Me', bbob.Gallagher101Me))
  def testNumpyExperimenter(self, func):
    dim = 3
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))

    # Asserts parameters are the same.
    parameters = list(exptr.problem_statement().search_space.parameters)
    self.assertLen(parameters, dim)

    discretization = {
        parameters[0].name: ['-1', '0', '1'],
        parameters[1].name: [0, 1, 2]
    }

    dis_exptr = discretizing_experimenter.DiscretizingExperimenter(
        exptr, discretization)
    discretized_parameters = dis_exptr.problem_statement(
    ).search_space.parameters

    self.assertLen(discretized_parameters, dim)
    self.assertListEqual([p.type for p in discretized_parameters], [
        pyvizier.ParameterType.CATEGORICAL, pyvizier.ParameterType.DISCRETE,
        pyvizier.ParameterType.DOUBLE
    ])

    parameters = {
        parameters[0].name: '0',
        parameters[1].name: 1,
        parameters[2].name: 1.5
    }
    t = pyvizier.Trial(parameters=parameters)

    dis_exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name
    self.assertAlmostEqual(
        func(np.array([0.0, 1.0, 1.5])),
        t.final_measurement.metrics[metric_name].value)
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)
    self.assertDictEqual(t.parameters.as_dict(), parameters)

  def testGridCreation(self):
    dim = 3
    func = bbob.Sphere
    problem_statement = bbob.DefaultBBOBProblemStatement(dim)
    # Mutate the last parameter.
    parameters = list(problem_statement.search_space.parameters)
    log_param = problem_statement.search_space.pop(parameters[-1].name)
    problem_statement.search_space.add(
        pyvizier.ParameterConfig.factory(
            log_param.name,
            scale_type=pyvizier.ScaleType.LOG,
            bounds=(0.01, 10.0),
        )
    )
    exptr = numpy_experimenter.NumpyExperimenter(func, problem_statement)

    parameters = list(exptr.problem_statement().search_space.parameters)
    self.assertLen(parameters, dim)

    discretization = {parameters[0].name: 3, parameters[-1].name: 4}

    dis_exptr = (
        discretizing_experimenter.DiscretizingExperimenter.create_with_grid(
            exptr, discretization
        )
    )
    search_space = dis_exptr.problem_statement().search_space
    self.assertEqual(search_space.num_parameters(), dim)
    self.assertEqual(
        search_space.num_parameters(pyvizier.ParameterType.DISCRETE), 2
    )
    self.assertLen(search_space.parameters[0].feasible_values, 3)
    self.assertSequenceEqual(
        search_space.parameters[0].feasible_values, [-5, 0, 5]
    )
    self.assertLen(search_space.parameters[-1].feasible_values, 4)
    self.assertSequenceAlmostEqual(
        search_space.parameters[-1].feasible_values,
        [0.01, 0.1, 1.0, 10.0],
        places=5,
    )

  def testGridCreationError(self):
    dim = 3
    func = bbob.Sphere
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim)
    )
    discretization = {'not_found_error': 3}

    with self.assertRaisesRegex(ValueError, 'not in search space'):
      discretizing_experimenter.DiscretizingExperimenter.create_with_grid(
          exptr, discretization
      )

  def testGridCreationErrorNonDouble(self):
    dim = 5
    func = bbob.Sphere
    problem_statement = bbob.DefaultBBOBProblemStatement(dim)
    exptr = numpy_experimenter.NumpyExperimenter(func, problem_statement)
    parameters = list(exptr.problem_statement().search_space.parameters)
    discretization = {parameters[0].name: 3, parameters[1].name: 4}

    dis_exptr = (
        discretizing_experimenter.DiscretizingExperimenter.create_with_grid(
            exptr, discretization
        )
    )

    with self.assertRaisesRegex(ValueError, 'Non-double parameters'):
      discretizing_experimenter.DiscretizingExperimenter.create_with_grid(
          dis_exptr, discretization
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for numpy_experimenter."""

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class NumpyExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere), ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated), ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar), ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass), ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel), ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek), ('Gallagher101Me', bbob.Gallagher101Me))
  def testNumpyExperimenter(self, func):
    dim = 2
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })

    exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name
    self.assertAlmostEqual(
        func(np.array([0.0, 1.0])),
        t.final_measurement.metrics[metric_name].value)
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)

  def testNonFinite(self):
    dim = 2
    exptr = numpy_experimenter.NumpyExperimenter(
        impl=lambda x: np.inf,
        problem_statement=bbob.DefaultBBOBProblemStatement(dim))

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t1 = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })
    t2 = pyvizier.Trial(parameters={
        param.name: -float(index) for index, param in enumerate(parameters)
    })

    trials = [t1, t2]
    exptr.evaluate(trials)
    for trial in trials:
      self.assertEmpty(trial.final_measurement.metrics)
      self.assertTrue(trial.infeasible)

  def testNotInSearchSpace(self):
    exptr = numpy_experimenter.NumpyExperimenter(
        impl=lambda x: x,
        problem_statement=bbob.DefaultBBOBProblemStatement(1),
    )

    t1 = pyvizier.Trial(parameters={'yyyy': 0.0})
    with self.assertRaises(ValueError):
      exptr.evaluate([t1])


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Meta learning helper functionality."""

import attrs
from vizier import pyvizier as vz


@attrs.define(kw_only=True)
class MetaLearningUtils:
  """Meta learning utils class."""

  # The objective goal of 'meta' and 'tuned' algorithms.
  _goal: vz.ObjectiveMetricGoal
  # The metric name associated with the 'tuned' designer trials.
  _tuned_metric_name: str
  # The metric name associated with the 'meta' designer trials.
  _meta_metric_name: str
  # The 'tuned' algorithms params to be tuned (meta designer's search space).
  _tuning_params: vz.SearchSpace

  @property
  def meta_problem(self) -> vz.ProblemStatement:
    """Create meta problem."""
    problem = vz.ProblemStatement(search_space=self._tuning_params)
    problem.metric_information = vz.MetricsConfig(
        metrics=[vz.MetricInformation(self._meta_metric_name, goal=self._goal)]
    )
    return problem

  def complete_meta_suggestion(
      self,
      meta_suggestion: vz.TrialSuggestion,
      score: float,
  ) -> vz.Trial:
    """Complete meta trial suggestion and assign score."""
    meta_trial = meta_suggestion.to_trial()
    meta_trial.complete(vz.Measurement(metrics={self._meta_metric_name: score}))
    return meta_trial

  def get_default_hyperparameters(self) -> vz.TrialSuggestion:
    """Generate trial suggestion with default tuned designer parameter values."""
    suggestion = vz.TrialSuggestion()
    for param in self.meta_problem.search_space.parameters:
      if param.default_value is None:
        raise ValueError(
            "Hyper-param (%s) doesn't have default value!" % param.name
        )
      else:
        suggestion.parameters[param.name] = param.default_value
    return suggestion

  def get_best_meta_trial(self, trials: list[vz.Trial]) -> vz.Trial:
    """Return the best meta trial."""
    return self._get_best_trial(trials, self._meta_metric_name)

  def get_best_tuned_trial(self, trials: list[vz.Trial]) -> vz.Trial:
    """Return the best tuned trial."""
    return self._get_best_trial(trials, self._tuned_metric_name)

  def get_best_meta_trial_score(self, trials: list[vz.Trial]) -> float:
    """Return the best meta trial score."""
    return self._get_best_trial_score(trials, self._meta_metric_name)

  def get_best_tuned_trial_score(self, trials: list[vz.Trial]) -> float:
    """Return the best tuned trial score."""
    return self._get_best_trial_score(trials, self._tuned_metric_name)

  def _get_best_trial(
      self, trials: list[vz.Trial], metric_name: str
  ) -> vz.Trial:
    """Return the best trial with the metric specified by `trial_type`."""
    best_trial = trials[0]
    for trial in trials:
      if self._is_trial_better(trial, best_trial, metric_name):
        best_trial = trial
    return best_trial

  def _get_best_trial_score(
      self, trials: list[vz.Trial], metric_name: str
  ) -> float:
    """Return the best trial score."""
    best_trial = self._get_best_trial(trials, metric_name)
    if best_trial.final_measurement is None:
      raise ValueError("'final_measurement' shouldn't be None.")
    return best_trial.final_measurement.metrics[metric_name].value

  def _is_trial_better(
      self,
      trial1: vz.Trial,
      trial2: vz.Trial,
      metric_name: str,
  ) -> bool:
    """Returns whether trial1 is better than trial2 in terms of `metric_name`."""
    if metric_name not in trial1.final_measurement_or_die.metrics:
      raise ValueError(
          f"Metric name ({metric_name}) not found in trial1. {trial1}"
      )
    if metric_name not in trial2.final_measurement_or_die.metrics:
      raise ValueError(
          f"Metric name ({metric_name}) not found in trial2. {trial2}"
      )
    if trial1.infeasible:
      return False
    if trial2.infeasible:
      return True
    val1 = trial1.final_measurement_or_die.metrics[metric_name].value
    val2 = trial2.final_measurement_or_die.metrics[metric_name].value
    if self._goal == vz.ObjectiveMetricGoal.MAXIMIZE and val1 > val2:
      return True
    elif self._goal == vz.ObjectiveMetricGoal.MINIMIZE and val1 < val2:
      return True
    else:
      return False



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests on core.py."""


import chex
from jax import numpy as jnp
from vizier._src.jax.optimizers import core
from absl.testing import absltest


class GetBestParamsTest(absltest.TestCase):

  def test_best_n_none(self):
    losses = jnp.array([
        3.0,
        1.0,
        4.0,
        2.0,
        5.0,
    ])
    all_params = {
        'a': losses + 0.1,
        'b': jnp.stack([losses + 0.2, losses + 0.3], axis=1),
    }
    actual = core.get_best_params(losses, all_params, best_n=None)
    expected = {'a': jnp.array(1.1), 'b': jnp.array([1.2, 1.3])}

    chex.assert_trees_all_close(actual, expected)

  def test_best_n_2(self):
    losses = jnp.array([
        3.0,
        1.0,
        4.0,
        2.0,
        5.0,
    ])
    all_params = {
        'a': losses + 0.1,
        'b': jnp.stack([losses + 0.2, losses + 0.3], axis=1),
    }
    actual = core.get_best_params(losses, all_params, best_n=2)
    expected = {
        'a': jnp.array([1.1, 2.1]),
        'b': jnp.array([[1.2, 1.3], [2.2, 2.3]]),
    }

    chex.assert_trees_all_close(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for stochastic_process_model."""

import functools
from typing import Optional
from unittest import mock

from absl import logging
from absl.testing import parameterized
import equinox as eqx
from flax import linen as nn
import jax
from jax import config
from jax import numpy as jnp
from jax import random
import numpy as np
from tensorflow_probability.substrates import jax as tfp
import tree
from vizier._src.jax import stochastic_process_model as sp_model
from vizier._src.jax import types
from vizier._src.jax.models import mask_features

from absl.testing import absltest


tfb = tfp.bijectors
tfd = tfp.distributions
tfpk = tfp.math.psd_kernels
tfde = tfp.experimental.distributions
tfpke = tfp.experimental.psd_kernels


def _kernel_coroutine(dtype=np.float64):
  amplitude = yield sp_model.ModelParameter(
      init_fn=lambda k: random.exponential(k, dtype=dtype),
      constraint=sp_model.Constraint(bounds=(np.zeros([], dtype=dtype), None)),
      regularizer=lambda x: 1e-3 * x**2,
      name='amplitude',
  )
  one = np.array([1.0]).astype(dtype)
  inverse_length_scale_continuous = yield sp_model.ModelParameter.from_prior(
      tfd.Exponential(rate=one, name='inverse_length_scale_continuous')
  )
  inverse_length_scale_categorical = yield sp_model.ModelParameter.from_prior(
      tfd.Exponential(rate=one, name='inverse_length_scale_categorical')
  )
  inverse_length_scale = tfpke.ContinuousAndCategoricalValues(
      inverse_length_scale_continuous, inverse_length_scale_categorical
  )
  kernel = tfpk.ExponentiatedQuadratic(amplitude=amplitude, validate_args=True)
  return tfpke.FeatureScaledWithCategorical(
      kernel, inverse_length_scale, validate_args=True
  )


def _test_coroutine(
    inputs: Optional[types.ModelInput] = None,
    num_tasks=1,
    dtype=np.float64,
):
  """A coroutine that follows the `ModelCoroutine` protocol."""
  kernel = yield from _kernel_coroutine(dtype=dtype)
  if inputs is not None:
    kernel = mask_features.MaskFeatures(
        kernel,
        dimension_is_missing=tfpke.ContinuousAndCategoricalValues(
            continuous=inputs.continuous.is_missing[1],
            categorical=inputs.categorical.is_missing[1],
        ),
    )
    inputs = tfpke.ContinuousAndCategoricalValues(
        inputs.continuous.padded_array, inputs.categorical.padded_array
    )
  if num_tasks == 1:
    return tfd.StudentTProcess(
        df=np.array(5.0).astype(dtype),
        kernel=kernel,
        index_points=inputs,
        observation_noise_variance=np.ones([], dtype=dtype),
        validate_args=True,
    )

  multi_task_kernel = tfpke.Independent(num_tasks=num_tasks, base_kernel=kernel)
  return tfde.MultiTaskGaussianProcess(
      kernel=multi_task_kernel,
      index_points=inputs,
      observation_noise_variance=np.ones([], dtype=dtype),
      validate_args=True,
  )


def _make_inputs(
    key,
    dtype=jnp.float64,
    num_continuous=5,
    num_categorical=3,
    num_tasks=1,
    num_observed=15,
    num_predictive=8,
    predictive_batch=(),
    pad_obs=0,
    pad_continuous_dim=0,
    pad_categorical_dim=0,
    pad_tasks=0,
):
  cont_obs_key, cat_obs_key, cont_pred_key, cat_pred_key, y_key = random.split(
      key, num=5
  )
  x_observed_cont = random.uniform(
      cont_obs_key, shape=(num_observed, num_continuous), dtype=dtype
  )
  x_observed_cat = random.randint(
      cat_obs_key, shape=(num_observed, num_categorical), minval=0, maxval=6
  )
  x_observed = types.ModelInput(
      continuous=types.PaddedArray.from_array(
          x_observed_cont,
          (num_observed + pad_obs, num_continuous + pad_continuous_dim),
          fill_value=np.nan,
      ),
      categorical=types.PaddedArray.from_array(
          x_observed_cat,
          (num_observed + pad_obs, num_categorical + pad_categorical_dim),
          fill_value=-1,
      ),
  )

  y_observed = random.uniform(
      y_key, shape=(num_observed, num_tasks), dtype=dtype
  )
  y_observed = types.PaddedArray.from_array(
      y_observed,
      (num_observed + pad_obs, num_tasks + pad_tasks),
      fill_value=np.nan,
  )

  x_predictive_cont = random.uniform(
      cont_pred_key,
      shape=predictive_batch + (num_predictive, num_continuous),
      dtype=dtype,
  )
  x_predictive_cat = random.randint(
      cat_pred_key,
      shape=predictive_batch + (num_predictive, num_categorical),
      minval=0,
      maxval=6,
  )

  x_predictive = types.ModelInput(
      continuous=types.PaddedArray.from_array(
          x_predictive_cont,
          predictive_batch
          + (num_predictive + pad_obs, num_continuous + pad_continuous_dim),
          fill_value=np.nan,
      ),
      categorical=types.PaddedArray.from_array(
          x_predictive_cat,
          predictive_batch
          + (num_predictive + pad_obs, num_categorical + pad_categorical_dim),
          fill_value=-1,
      ),
  )
  return x_observed, y_observed, x_predictive


class MeanFn(nn.Module):

  @nn.compact
  def __call__(self, x):
    x = nn.relu(nn.Dense(3)(x.continuous.padded_array))
    x = nn.Dense(1)(x)
    return jnp.squeeze(x, axis=-1)


class StochasticProcessModelTest(parameterized.TestCase):

  @parameterized.named_parameters(
      # TODO: Fix support for f32.
      {
          'testcase_name': 'continuous_only',
          'input_kwargs': dict(
              num_categorical=0, pad_obs=3, pad_continuous_dim=2
          ),
      },
      {
          'testcase_name': 'multitask_continuous',
          'input_kwargs': dict(
              num_categorical=2,
              num_tasks=3,
          ),
      },
      {
          'testcase_name': 'continuous_categorical',
          'input_kwargs': dict(
              pad_continuous_dim=5, pad_categorical_dim=4, pad_obs=7
          ),
      },
      {'testcase_name': 'continuous_categorical_nopad', 'input_kwargs': dict()},
      # TODO: Enable `is_missing` in TFP Multitask GP.
      # {
      #     'testcase_name': 'continuous_categorical_multitask',
      #     'input_kwargs': dict(
      #         num_tasks=3,
      #         pad_tasks=2,
      #         pad_obs=5,
      #     )
      # },
  )
  def test_stochastic_process_model(self, input_kwargs):
    init_key, data_key, sample_key = jax.random.split(random.PRNGKey(0), num=3)
    x_observed, y_observed, x_predictive = _make_inputs(
        data_key, **input_kwargs
    )
    dtype = y_observed.padded_array.dtype
    model_coroutine = functools.partial(
        _test_coroutine,
        num_tasks=input_kwargs.get('num_tasks', 1),
        dtype=dtype,
    )
    model = sp_model.StochasticProcessModel(coroutine=model_coroutine)

    init_state = model.init(init_key, x_observed)
    dist, losses = model.apply(init_state, x_observed, mutable=('losses',))
    if isinstance(dist, tfde.MultiTaskGaussianProcess):
      lp = dist.log_prob(y_observed.padded_array)
    else:
      lp = dist.log_prob(
          y_observed.padded_array[:, 0], is_missing=y_observed.is_missing[0]
      )
    self.assertTrue(np.isfinite(lp))
    self.assertEqual(lp.dtype, dtype)

    # Check regularization loss values.
    gen = model_coroutine()
    p = next(gen)
    params = dict(init_state['params'])
    losses = dict(losses['losses'])
    try:
      while True:
        value = params[p.name]
        param_loss = losses[f'{p.name}_regularization']
        self.assertTrue(np.isfinite(value))
        self.assertEqual(value.dtype, dtype)
        self.assertTrue(np.isfinite(param_loss))
        self.assertAlmostEqual(p.regularizer(value), param_loss)
        p = gen.send(value)
    except StopIteration:
      pass

    _, pp_state = model.apply({'params': params},
                              x_observed,
                              y_observed,
                              method=model.precompute_predictive,
                              mutable=('predictive',))

    state = {'params': params, **pp_state}
    x_pred = types.ModelInput(
        continuous=x_predictive.continuous.replace_fill_value(0.0),
        categorical=x_predictive.categorical.replace_fill_value(0),
    )
    pp_dist = model.apply(
        state,
        x_pred,
        x_observed,
        y_observed,
        method=model.posterior_predictive,
    )
    pp_log_prob = pp_dist.log_prob(pp_dist.sample(seed=sample_key))
    self.assertTrue(np.isfinite(pp_log_prob).all())
    self.assertEqual(pp_log_prob.dtype, dtype)

  @parameterized.named_parameters(
      {
          'testcase_name': 'continuous_with_mean_fn',
          'num_tasks': 1,
          'use_mean_fn': True,
      },
      {
          'testcase_name': 'multitask_continuous',
          'num_tasks': 3,
      },
      {
          'testcase_name': 'multitask_continuous_categorical',
          'num_tasks': 3,
      },
  )
  def test_jax_transformations(self, num_tasks, use_mean_fn=False):
    init_key, data_key, prior_key, post_key = jax.random.split(
        random.PRNGKey(0), num=4
    )
    num_observed = 20
    x_observed, y_observed, x_predictive = _make_inputs(
        data_key,
        num_tasks=num_tasks,
        num_observed=num_observed,
        predictive_batch=(12,),
        dtype=np.float32,
    )
    model = sp_model.StochasticProcessModel(
        coroutine=functools.partial(
            _test_coroutine,
            num_tasks=num_tasks,
            dtype=np.float32,
        ),
        mean_fn=MeanFn() if use_mean_fn else None,
    )

    batch_size = 10
    keys = jax.random.split(init_key, num=batch_size)
    init_state = jax.vmap(lambda k: model.init(k, x_observed))(keys)
    s = jax.jit(
        jax.vmap(
            lambda s: model.apply(  # pylint: disable=g-long-lambda
                s, x_observed, mutable=('losses',)
            )[0].sample(seed=prior_key)
        )
    )(init_state)
    sample_shape = (
        (batch_size, num_observed)
        if num_tasks == 1
        else (batch_size, num_observed, num_tasks)
    )
    self.assertEqual(s.shape, sample_shape)

    _, pp_state = jax.jit(
        jax.vmap(
            lambda p: model.apply(  # pylint: disable=g-long-lambda
                p,
                x_observed,
                y_observed,
                method=model.precompute_predictive,
                mutable=('predictive',),
            )
        )
    )(init_state)

    state = {**init_state, **pp_state}

    @jax.jit
    @jax.vmap
    def _posterior_sample_and_log_prob(s):
      dist = model.apply(
          s,
          x_predictive,
          x_observed,
          y_observed,
          method=model.posterior_predictive,
      )[0]
      sample = dist.sample(seed=post_key)
      return dist.log_prob(sample)

    pp_log_prob = _posterior_sample_and_log_prob(state)
    log_prob_grad = jax.grad(
        lambda s: jnp.sum(_posterior_sample_and_log_prob(s))
    )(state)
    for g in tree.flatten(log_prob_grad['params']):
      self.assertTrue(np.isfinite(g).all())
    self.assertTrue(np.isfinite(pp_log_prob).all())
    self.assertEqual(pp_log_prob.shape, (batch_size,))

  def test_cholesky_not_recomputed(self):
    # Need num_observed == num_predictive so the mock Cholesky works for both
    # observed and predictive.
    x_obs, y_obs, x = _make_inputs(
        jax.random.PRNGKey(5), np.float32, num_observed=12, num_predictive=12
    )
    chol = np.eye(12, dtype=np.float32)
    mock_cholesky_fn = mock.Mock(return_value=chol)

    def _coro_with_mock_cholesky(inputs=None):
      stp = yield from _test_coroutine(inputs=inputs, dtype=np.float32)
      # The `copy` method of TFP distributions rebuilds the distribution object
      # with the given constructor kwarg(s) replaced.
      return stp.copy(cholesky_fn=mock_cholesky_fn)

    model = sp_model.StochasticProcessModel(coroutine=_coro_with_mock_cholesky)
    keys = jax.random.split(jax.random.PRNGKey(0), num=8)
    params = jax.vmap(lambda k: model.init(k, x_obs))(keys)
    _, pp_state = jax.vmap(
        lambda p: model.apply(  # pylint: disable=g-long-lambda
            p,
            x_obs,
            y_obs,
            method=model.precompute_predictive,
            mutable=('predictive',),
        )
    )(params)
    mock_cholesky_fn.assert_called_once()

    # Sampling from the posterior predictive should invoke cholesky_fn only
    # once (on the kernel matrix for the predictive index points).
    mock_cholesky_fn.reset_mock()
    state = {**params, **pp_state}
    _ = jax.jit(
        jax.vmap(
            lambda s: model.apply(  # pylint: disable=g-long-lambda
                s, x, x_obs, y_obs, method=model.posterior_predictive
            ).sample(seed=jax.random.PRNGKey(0))
        )
    )(state)
    mock_cholesky_fn.assert_called_once()

    # Sample new data and compute the posterior predictive.
    mock_cholesky_fn.reset_mock()
    x_obs2, y_obs2, x2 = _make_inputs(
        jax.random.PRNGKey(6), np.float32, num_observed=12, num_predictive=12
    )
    params2 = model.init(jax.random.PRNGKey(7), x_obs2)
    _, pp_state2 = model.apply(
        params2,
        x_obs2,
        y_obs2,
        method=model.precompute_predictive,
        mutable=('predictive',),
    )
    _ = jax.jit(
        lambda s: model.apply(  # pylint: disable=g-long-lambda
            s, x2, x_obs2, y_obs2, method=model.posterior_predictive
        ).sample(seed=jax.random.PRNGKey(0))
    )({**params2, **pp_state2})
    # Two Cholesky calls; one for observed, one for predictive.
    self.assertEqual(mock_cholesky_fn.call_count, 2)

    # Assert that `model.posterior_predictive`, when called on the original
    # dataset, does not recompute the Cholesky.
    mock_cholesky_fn.reset_mock()
    state = {**params, **pp_state}
    _ = jax.jit(
        jax.vmap(
            lambda s: model.apply(  # pylint: disable=g-long-lambda
                s, x, x_obs, y_obs, method=model.posterior_predictive
            ).sample(seed=jax.random.PRNGKey(0))
        )
    )(state)
    mock_cholesky_fn.assert_called_once()

  def test_stochastic_process_model_with_mean_fn(self):

    data_key, init_key, vmap_key = random.split(random.PRNGKey(0), num=3)
    x_observed, y_observed, x_predictive = _make_inputs(
        data_key,
        num_continuous=3,
        num_categorical=0,
        num_observed=20,
        num_predictive=50,
    )

    # Use `jnp.squeeze` to remove the singleton dimension of the output of
    # `mean_fn`, so that it has shape `(num_observed,)`.
    mean_fn = MeanFn()
    model = sp_model.StochasticProcessModel(
        coroutine=_test_coroutine, mean_fn=mean_fn)

    init_state = model.init(init_key, x_observed)
    stp = model.apply(init_state, x_observed, mutable=False)
    lp = stp.log_prob(y_observed.padded_array[:, 0])
    self.assertTrue(np.isfinite(lp))

    # The mean of the GP should equal `mean_fn` evaluated at the index points.
    np.testing.assert_allclose(
        stp.mean(),
        mean_fn.apply({'params': init_state['params']['mean_fn']}, x_observed))

    _, pp_state = model.apply({'params': init_state['params']},
                              x_observed,
                              y_observed,
                              method=model.precompute_predictive,
                              mutable=('predictive',))

    state = {'params': init_state['params'], **pp_state}
    pp_dist = model.apply(
        state,
        x_predictive,
        x_observed,
        y_observed,
        method=model.posterior_predictive,
    )
    pp_mean = pp_dist.mean()

    self.assertTrue(
        np.isfinite(
            pp_dist.log_prob(x_predictive.continuous.padded_array.sum(axis=-1))
        ).all()
    )
    self.assertSequenceEqual(pp_mean.shape, (50,))
    self.assertTrue(np.isfinite(pp_mean).all())

    # Test that vmap works.
    def log_prob(p):
      return model.apply(p, x_observed, mutable=('losses',))[0].log_prob(
          y_observed.padded_array
      )

    batch_size = 20
    keys = random.split(vmap_key, num=batch_size)
    params = jax.vmap(lambda k: model.init(k, x_observed))(keys)
    lp = jax.vmap(log_prob)(params)
    self.assertLen(lp, batch_size)
    self.assertTrue(np.isfinite(lp).all())


class VectorToArrayTreeTest(absltest.TestCase):

  def test_vector_to_array_tree(self):
    k0, k1 = random.split(random.PRNGKey(0))
    params = {
        'foo': random.uniform(k0, shape=(2, 3)),
        'bar': np.array(0.3),
        'baz': random.uniform(k1, shape=(6,))
    }
    bijector = sp_model.VectorToArrayTree(params)
    self.assertEqual(bijector.inverse(params).shape, (13,))

    v = np.ones([13])
    struct = bijector.forward(v)
    self.assertSameElements(list(struct.keys()), list(params.keys()))
    for k in params:
      self.assertEqual(struct[k].shape, params[k].shape)


class ModelParameterTest(absltest.TestCase):

  def test_parameter_from_prior(self):
    prior = tfd.Gamma(concentration=1., rate=1.)
    param = sp_model.ModelParameter.from_prior(prior=prior)
    samples = param.init_fn(random.PRNGKey(0))
    regularization = param.regularizer(samples)
    x = np.linspace(-10.0, 10.0, 20)
    self.assertTrue(np.isfinite(regularization))
    self.assertEmpty(regularization.shape)
    self.assertNotEqual(regularization, 0.)
    self.assertTrue((param.constraint.bijector(x) > 0.0).all())


class ConstraintTest(parameterized.TestCase):

  @parameterized.parameters(
      (-1.0, None),
      (0.0, 10.0),
      (None, 1.0),
      ({'a': -1.0, 'b': 0.0}, {'a': 1.0, 'b': 10.0}),
      ({'a': None, 'b': 0.0}, {'a': 1.0, 'b': None}),
      (None, {'a': 1.0, 'b': 2.0}),
  )
  def test_create_constraint(self, lower, upper):
    make_array = lambda x: None if x is None else jnp.array(x)
    lower = jax.tree_util.tree_map(make_array, lower)
    upper = jax.tree_util.tree_map(make_array, upper)
    constraint = sp_model.Constraint.create(
        bounds=(lower, upper), bijector_fn=tfb.SoftClip
    )
    x_part = jnp.linspace(-5.0, 5.0, 10)
    x = tree.map_structure(lambda _: x_part, lower or upper)
    y = constraint.bijector(x)
    for y_, b_ in zip(tree.flatten(y), tree.flatten(lower)):
      if b_ is not None:
        self.assertTrue((y_ > b_).all())
    for y_, b_ in zip(tree.flatten(y), tree.flatten(upper)):
      if b_ is not None:
        self.assertTrue((y_ < b_).all())

  @parameterized.parameters((None,), (MeanFn,))
  def test_get_constraints(self, mean_fn):
    if mean_fn:
      mean_fn = mean_fn()

    x, *_ = _make_inputs(jax.random.PRNGKey(0))
    model = sp_model.StochasticProcessModel(
        coroutine=_test_coroutine, mean_fn=mean_fn
    )

    constraint = sp_model.get_constraints(model, x=x)
    unconstrained_p = model.lazy_init(jax.random.PRNGKey(0), x)['params']
    p = constraint.bijector(unconstrained_p)
    lower, upper = constraint.bounds

    tree.assert_same_structure(p, unconstrained_p)
    for y_, b_ in zip(tree.flatten(p), tree.flatten(lower)):
      if b_ is not None:
        self.assertTrue((y_ >= b_).all())
    for y_, b_ in zip(tree.flatten(p), tree.flatten(upper)):
      if b_ is not None:
        self.assertTrue((y_ <= b_).all())


@functools.lru_cache(maxsize=None)
def _test_data() -> tuple[types.ModelData, types.ModelInput]:
  x_observed, y_observed, x_predictive = _make_inputs(
      jax.random.PRNGKey(0),
      num_continuous=5,
      num_categorical=2,
      num_observed=13,
      num_predictive=7,
      pad_continuous_dim=2,
      pad_categorical_dim=3,
  )
  return types.ModelData(features=x_observed, labels=y_observed), x_predictive


class EquinoxModulesTest(absltest.TestCase):

  def testCoroutineWithData(self):
    data = _test_data()[0]
    data = types.ModelData(
        features=types.ModelInput(
            data.features.continuous.replace_fill_value(0.0),
            data.features.categorical.replace_fill_value(0),
        ),
        labels=data.labels,
    )
    wrapper = sp_model.CoroutineWithData(_test_coroutine, data)

    # Test setup and loss
    params = wrapper.setup(jax.random.PRNGKey(0))
    self.assertTrue(np.all(np.isfinite(wrapper.loss_with_aux(params)[0])))

    # Test constraint
    constraint = wrapper.constraints()
    p = constraint.bijector(params)
    lower, upper = constraint.bounds

    tree.assert_same_structure(p, params)
    for y_, b_ in zip(tree.flatten(p), tree.flatten(lower)):
      if b_ is not None:
        self.assertTrue((y_ >= b_).all())
    for y_, b_ in zip(tree.flatten(p), tree.flatten(upper)):
      if b_ is not None:
        self.assertTrue((y_ <= b_).all())

  def testStochasticProcessWithCoroutine(self):
    model = sp_model.StochasticProcessWithCoroutine.initialize(
        _test_coroutine, rng=jax.random.PRNGKey(0)
    )
    dist, _ = model.call_with_aux(_test_data()[0].features)
    self.assertSequenceEqual(dist.event_shape, (13,))
    self.assertSequenceEqual(dist.batch_shape, tuple())

    data = _test_data()[0]
    data = types.ModelData(
        features=types.ModelInput(
            data.features.continuous.replace_fill_value(0.0),
            data.features.categorical.replace_fill_value(0),
        ),
        labels=data.labels,
    )
    loss, aux = model.loss_with_aux(data)
    logging.info('%s, %s', loss, aux)
    logging.info('model:%s', model.params)
    self.assertTrue(np.isfinite(loss))

  def testPrecomputePredictive(self):
    model = sp_model.StochasticProcessWithCoroutine.initialize(
        _test_coroutine, rng=jax.random.PRNGKey(0)
    )
    predictive = model.precompute_predictive(_test_data()[0])
    dist = predictive.predict(_test_data()[1])
    self.assertSequenceEqual(dist.event_shape, (7,))
    self.assertSequenceEqual(dist.batch_shape, tuple())

    x_obs_empty, y_obs_empty, y_pred = _make_inputs(
        key=jax.random.PRNGKey(0), num_observed=0, num_predictive=7
    )
    predictive_empty_obs = model.precompute_predictive(
        types.ModelData(x_obs_empty, y_obs_empty)
    )
    dist = predictive_empty_obs.predict(y_pred)
    self.assertSequenceEqual(dist.event_shape, (7,))
    self.assertSequenceEqual(dist.batch_shape, tuple())


class UniformEnsemblePrecomputePredictiveTest(parameterized.TestCase):

  def test_no_batch_shape(self):
    model = sp_model.StochasticProcessWithCoroutine.initialize(
        _test_coroutine, rng=jax.random.PRNGKey(0)
    )
    predictive = model.precompute_predictive(_test_data()[0])
    dist = predictive.predict(_test_data()[1])
    self.assertSequenceEqual(dist.event_shape, (7,))
    self.assertSequenceEqual(dist.batch_shape, tuple())

  @parameterized.parameters(dict(n=5), dict(n=1))
  def test_batch_shape_n(self, n):
    model = jax.vmap(
        eqx.Partial(
            sp_model.StochasticProcessWithCoroutine.initialize,
            coroutine=_test_coroutine,
        )
    )(rng=jax.random.split(jax.random.PRNGKey(0), n))

    predictive = model.precompute_predictive(_test_data()[0])
    dist = predictive.predict(_test_data()[1])
    self.assertSequenceEqual(dist.event_shape, (7,))
    self.assertSequenceEqual(dist.batch_shape, (n,))

    # Put it in the UniformEnsemblePredictive and the batch dimension is gone.
    ensemble = sp_model.UniformEnsemblePredictive(predictive)
    dist = ensemble.predict(_test_data()[1])
    self.assertSequenceEqual(dist.event_shape, (7,))
    self.assertSequenceEqual(dist.batch_shape, tuple())


if __name__ == '__main__':
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier import algorithms as vza
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.runners import benchmark_state

from absl.testing import absltest


class BenchmarkStateTest(absltest.TestCase):

  def test_policy_suggester_active_trials_with_init(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('x', 0.0, 1.0)
    problem.search_space.root.add_float_param('y', 0.0, 1.0)
    problem.metric_information.append(
        vz.MetricInformation(
            name='maximize_metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    # Set up policy supporter with one Active and Completed Trial.
    policy_supporter = pythia.InRamPolicySupporter(problem)
    trials = [
        vz.Trial(parameters={'x': 0.5, 'y': 0.6}).complete(
            vz.Measurement(metrics={'maximize_metric': 0.3})
        ),
        vz.Trial(parameters={'x': 0.1, 'y': 0.2}),
    ]
    # The Completed and Active Trials should be updated in Designer.
    policy_supporter.AddTrials(trials)

    class DummyDesigner(vza.Designer):

      def __init__(self, problem_statement):
        self._designer = random.RandomDesigner(problem_statement.search_space)
        self.completed = []
        self.all_active = []

      def suggest(self, count):
        return self._designer.suggest(count)

      def update(self, completed, all_active):
        self.completed.extend(completed.trials)
        self.all_active = all_active.trials

    designer = DummyDesigner(problem)
    suggester = benchmark_state.PolicySuggester.from_designer_factory(
        problem, lambda _, **kwargs: designer, supporter=policy_supporter
    )

    suggestions = list(suggester.suggest(batch_size=5))
    self.assertLen(designer.completed, 1)
    self.assertLen(designer.all_active, 1)
    self.assertLen(suggestions, 5)

    # Report COMPLETED Trial back to supporter directly.
    completed_trial = suggestions[0].complete(
        vz.Measurement(metrics={'maximize_metric': 0.4})
    )
    suggester.supporter.AddTrials([completed_trial])
    suggester.suggest(1)
    self.assertLen(designer.completed, 2)
    self.assertLen(designer.all_active, 5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for Eagle Strategy utils."""

import copy
from typing import Optional

import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy_utils
from vizier._src.algorithms.designers.eagle_strategy import testing
from absl.testing import absltest
from absl.testing import parameterized

EagleStrategyUtils = eagle_strategy_utils.EagleStrategyUtils
FireflyAlgorithmConfig = eagle_strategy_utils.FireflyAlgorithmConfig
EagleStrategyUtils = eagle_strategy_utils.EagleStrategyUtils
FireflyPool = eagle_strategy_utils.FireflyPool
Firefly = eagle_strategy_utils.Firefly


def _get_parameter_config(search_space: vz.SearchSpace,
                          param_name: str) -> Optional[vz.ParameterConfig]:
  """Iterates over the search space parameters to find parameter by name."""
  for param_config in search_space.parameters:
    if param_config.name == param_name:
      return param_config


class UtilsTest(parameterized.TestCase):
  """Tests for the EagleStrategyUtuls class."""

  def setUp(self):
    super(UtilsTest, self).setUp()
    self.rng = np.random.default_rng(seed=0)
    self.search_space = vz.SearchSpace()
    root = self.search_space.root
    root.add_bool_param('b1')
    root.add_discrete_param('d1', [1.0, 2.0, 9.0, 10.0])
    root.add_float_param('f1', 0.0, 15.0, scale_type=vz.ScaleType.LINEAR)
    root.add_float_param('f2', 0.0, 10.0, scale_type=vz.ScaleType.LINEAR)
    root.add_int_param('i1', 0, 10, scale_type=vz.ScaleType.LINEAR)
    root.add_categorical_param('c1', ['a', 'b', 'c'])

    self.param_dict1 = vz.ParameterDict()
    self.param_dict1['f1'] = 5.0
    self.param_dict1['f2'] = 1.0
    self.param_dict1['c1'] = 'a'
    self.param_dict1['d1'] = 10.0
    self.param_dict1['i1'] = 3
    self.param_dict1['b1'] = 'True'

    self.param_dict2 = vz.ParameterDict()
    self.param_dict2['f1'] = 2.0
    self.param_dict2['f2'] = 8.0
    self.param_dict2['c1'] = 'b'
    self.param_dict2['d1'] = 1.0
    self.param_dict2['i1'] = 1
    self.param_dict2['b1'] = 'True'

    metric_information_max = vz.MetricInformation(
        name='obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
    metric_information_min = vz.MetricInformation(
        name='obj', goal=vz.ObjectiveMetricGoal.MINIMIZE)

    self.problem_max = vz.ProblemStatement(
        search_space=self.search_space,
        metric_information=[metric_information_max])
    self.problem_min = vz.ProblemStatement(
        search_space=self.search_space,
        metric_information=[metric_information_min])

    self.utils = EagleStrategyUtils(self.problem_max, FireflyAlgorithmConfig(),
                                    self.rng)

  def test_compute_pool_capacity(self):
    factor = FireflyAlgorithmConfig().pool_size_factor
    expected_capacity = 10 + round((6**factor + 6) * 0.5)
    self.assertEqual(self.utils.compute_pool_capacity(), expected_capacity)

  def test_compute_cononical_distance(self):
    dist = self.utils.compute_cononical_distance(self.param_dict1,
                                                 self.param_dict2)
    self.assertIsInstance(dist, float)
    self.assertAlmostEqual(dist, 2.57)

  def test_compute_canonical_distance_squared_by_type(self):
    dists = self.utils._compute_canonical_distance_squared_by_type(
        self.param_dict1, self.param_dict2)
    self.assertEqual(dists[vz.ParameterType.CATEGORICAL], 1.0 + 0.0)
    self.assertEqual(dists[vz.ParameterType.INTEGER], ((3 - 1) / 10)**2)
    self.assertEqual(dists[vz.ParameterType.DOUBLE],
                     (3.0 / 15)**2 + (7.0 / 10)**2)
    self.assertEqual(dists[vz.ParameterType.DISCRETE], (9.0 / 9.0)**2)

  def test_is_better_than(self):
    trial1 = vz.Trial(parameters=self.param_dict1)
    trial1.complete(
        vz.Measurement(
            metrics={eagle_strategy_utils.OBJECTIVE_NAME: vz.Metric(
                value=2.0)}),
        inplace=True)
    trial2 = vz.Trial(parameters=self.param_dict2)
    trial2.complete(
        vz.Measurement(
            metrics={eagle_strategy_utils.OBJECTIVE_NAME: vz.Metric(
                value=1.5)}),
        inplace=True)
    # Test for maximization problem
    self.assertTrue(self.utils.is_better_than(trial1, trial2))
    self.assertFalse(self.utils.is_better_than(trial2, trial1))
    # Test for minimization problem
    min_problem_utils = copy.deepcopy(self.utils)
    min_problem_utils.problem_statement = self.problem_min
    min_problem_utils.__attrs_post_init__()
    self.assertFalse(min_problem_utils.is_better_than(trial1, trial2))
    self.assertTrue(min_problem_utils.is_better_than(trial2, trial1))

  def test_is_better_than_infeasible(self):
    trial1 = vz.Trial(parameters=self.param_dict1)
    trial1.complete(
        vz.Measurement(metrics={'obj': vz.Metric(value=2.0)}), inplace=True)
    trial1._infeasibility_reason = 'infeasible reason'
    trial2 = vz.Trial(parameters=self.param_dict2)
    trial2.complete(
        vz.Measurement(metrics={'obj': vz.Metric(value=1.5)}), inplace=True)
    self.assertFalse(self.utils.is_better_than(trial1, trial2))
    self.assertFalse(self.utils.is_better_than(trial2, trial1))

  def test_is_pure_categorical(self):
    pure_categorical_space = vz.SearchSpace()
    pure_categorical_space.root.add_bool_param('b1')
    pure_categorical_space.root.add_bool_param('b2')
    pure_categorical_space.root.add_categorical_param('c1', ['a', 'b'])
    pure_categorical_space.root.add_categorical_param('c2', ['a', 'b'])
    pure_categorical_space_utils = copy.deepcopy(self.utils)
    pure_categorical_space_utils._search_space = pure_categorical_space
    self.assertTrue(pure_categorical_space_utils.is_pure_categorical())
    self.assertFalse(self.utils.is_pure_categorical())

  def test_combine_two_parameters_integer(self):
    int_param_config = _get_parameter_config(self.search_space, 'i1')
    new_value = self.utils.combine_two_parameters(int_param_config,
                                                  self.param_dict1,
                                                  self.param_dict2, 0.1)
    self.assertEqual(new_value, round(3 * 0.1 + 1 * 0.9))

  def test_combine_two_parameters_float(self):
    float_param_config = _get_parameter_config(self.search_space, 'f1')
    new_value = self.utils.combine_two_parameters(float_param_config,
                                                  self.param_dict1,
                                                  self.param_dict2, 0.1)
    self.assertEqual(new_value, 5.0 * 0.1 + 2.0 * 0.9)

  def test_combine_two_parameters_discrete(self):
    float_param_config = _get_parameter_config(self.search_space, 'd1')
    new_value = self.utils.combine_two_parameters(float_param_config,
                                                  self.param_dict1,
                                                  self.param_dict2, 0.1)
    self.assertEqual(new_value, 2.0)

  @parameterized.named_parameters(
      dict(testcase_name='prob=0', prob=0.0, target='b'),
      dict(testcase_name='prob=1.0', prob=1.0, target='a'),
      dict(testcase_name='prob=1.5', prob=1.5, target='a'),
      dict(testcase_name='prob=-0.1', prob=-0.1, target='b'),
  )
  def test_combine_two_parameters_categorical1(self, prob, target):
    categorical_param_config = _get_parameter_config(self.search_space, 'c1')
    new_value = self.utils.combine_two_parameters(categorical_param_config,
                                                  self.param_dict1,
                                                  self.param_dict2, prob)
    self.assertEqual(new_value, target)

  def test_combine_two_parameters_categorical2(self):
    categorical_param_config = _get_parameter_config(self.search_space, 'c1')
    new_value = self.utils.combine_two_parameters(categorical_param_config,
                                                  self.param_dict1,
                                                  self.param_dict2, 0.5)
    self.assertIn(new_value, ['a', 'b'])

  @parameterized.named_parameters(
      dict(testcase_name='Above1', value=9.0, prob=0.999, target=10.0),
      dict(testcase_name='Above2', value=10.0, prob=0.999, target=10.0),
      dict(testcase_name='Below1', value=2.0, prob=-0.999, target=0.0),
      dict(testcase_name='Below2', value=0.0, prob=-0.999, target=0.0),
      dict(testcase_name='Change1', value=5.0, prob=0.212, target=7.12),
      dict(testcase_name='Change2', value=5.0, prob=-0.1, target=4.0),
      dict(testcase_name='Change3', value=2.0, prob=0.111, target=3.11),
  )
  def test_perturb_parameter_float(self, value, prob, target):
    decimal = vz.ParameterConfig.factory('f1', bounds=(0.0, 10.0))
    new_value = self.utils.perturb_parameter(decimal, value, prob)
    self.assertIsInstance(new_value, float)
    self.assertAlmostEqual(new_value, target)

  def test_perturb_parameter_categorical(self):
    categorical = vz.ParameterConfig.factory(
        'c1', feasible_values=['a', 'b', 'c'])
    new_value1 = self.utils.perturb_parameter(categorical, 'b', 0.2)
    self.assertIn(new_value1, ['a', 'b', 'c'])
    new_value2 = self.utils.perturb_parameter(categorical, 'b', 0.0)
    self.assertEqual(new_value2, 'b')

  @parameterized.named_parameters(
      dict(testcase_name='Above1', value=9, prob=0.999, target=10),
      dict(testcase_name='Above2', value=10, prob=0.999, target=10),
      dict(testcase_name='Below1', value=2, prob=-0.999, target=0),
      dict(testcase_name='Below2', value=0, prob=-0.999, target=0),
      dict(testcase_name='NoChange', value=2, prob=0.000001, target=2),
      dict(testcase_name='Change1', value=1, prob=0.09, target=2),
      dict(testcase_name='Change2', value=1, prob=0.51, target=6),
  )
  def test_perturb_parameter_integer(self, value, prob, target):
    integer = vz.ParameterConfig.factory('i1', bounds=(0, 10))
    new_value = self.utils.perturb_parameter(integer, value, prob)
    self.assertIsInstance(new_value, int)
    self.assertEqual(new_value, target)

  @parameterized.named_parameters(
      dict(testcase_name='Above1', value=9.0, prob=0.999, target=10.0),
      dict(testcase_name='Above2', value=10.0, prob=0.999, target=10.0),
      dict(testcase_name='Below1', value=2.0, prob=-0.999, target=1.0),
      dict(testcase_name='Below2', value=1.0, prob=-0.999, target=1.0),
      dict(testcase_name='NoChange', value=2.0, prob=0.000001, target=2.0),
      dict(testcase_name='Change', value=1.0, prob=0.09, target=2.0),
  )
  def test_perturb_parameter_discrete(self, value, prob, target):
    discrete = vz.ParameterConfig.factory(
        'd1', feasible_values=[1.0, 2.0, 9.0, 10.0])
    new_value = self.utils.perturb_parameter(discrete, value, prob)
    self.assertIsInstance(new_value, float)
    self.assertEqual(new_value, target)

  def test_degrees_of_freedom(self):
    dof = self.utils._degrees_of_freedom
    self.assertEqual(dof[vz.ParameterType.DOUBLE], 2)
    self.assertEqual(dof[vz.ParameterType.CATEGORICAL], 2)
    self.assertEqual(dof[vz.ParameterType.INTEGER], 1)
    self.assertEqual(dof[vz.ParameterType.DISCRETE], 1)

  def test_replace_trial_metric_name(self):
    search_space = vz.SearchSpace()
    root = search_space.root
    root.add_float_param('f1', 0.0, 15.0, scale_type=vz.ScaleType.LINEAR)
    metric_information = vz.MetricInformation(
        name='obj123', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
    problem = vz.ProblemStatement(
        search_space=search_space, metric_information=[metric_information])

    utils = EagleStrategyUtils(problem, FireflyAlgorithmConfig(), self.rng)
    metadata = vz.Metadata()
    metadata.ns('eagle')['parent_fly_id'] = '123'
    trial = vz.Trial(parameters={'f1': 0.0}, metadata=metadata)
    trial.complete(measurement=vz.Measurement(metrics={'obj123': 1123.3}))
    new_trial = utils.standardize_trial_metric_name(trial)
    self.assertEqual(new_trial.final_measurement.metrics['objective'].value,
                     1123.3)
    self.assertEqual(new_trial.parameters['f1'].value, 0.0)
    self.assertEqual(new_trial.metadata.ns('eagle')['parent_fly_id'], '123')


class FireflyPoolTest(absltest.TestCase):
  """Tests for the FireflyPool class."""

  def test_generate_new_fly_id(self):
    firefly_pool = testing.create_fake_empty_firefly_pool(capacity=2)
    self.assertEqual(firefly_pool.generate_new_fly_id(), 0)
    self.assertEqual(firefly_pool.generate_new_fly_id(), 1)
    self.assertEqual(firefly_pool.generate_new_fly_id(), 2)
    self.assertEqual(firefly_pool.generate_new_fly_id(), 3)

  def test_create_or_update_fly(self):
    # Test creating a new fly in the pool.
    firefly_pool = testing.create_fake_empty_firefly_pool()
    trial = testing.create_fake_trial(
        parent_fly_id=112, x_value=0, obj_value=0.8)
    firefly_pool.create_or_update_fly(trial, 112)
    self.assertEqual(firefly_pool.size, 1)
    self.assertLen(firefly_pool._pool, 1)
    self.assertIs(firefly_pool._pool[112].trial, trial)
    # Test that another trial with the same parent id updates the fly.
    trial2 = testing.create_fake_trial(
        parent_fly_id=112, x_value=1, obj_value=1.5)
    firefly_pool.create_or_update_fly(trial2, 112)
    self.assertEqual(firefly_pool.size, 1)
    self.assertLen(firefly_pool._pool, 1)
    self.assertIs(firefly_pool._pool[112].trial, trial2)

  def test_find_closest_parent(self):
    firefly_pool = testing.create_fake_populated_firefly_pool(
        x_values=[1, 2, 5], obj_values=[2, 10, -2], capacity=4)
    trial = testing.create_fake_trial(
        parent_fly_id=123, x_value=4.2, obj_value=8)
    parent_fly = firefly_pool.find_closest_parent(trial)
    self.assertEqual(parent_fly.id_, 2)

  def test_is_best_fly(self):
    firefly_pool = testing.create_fake_populated_firefly_pool(
        x_values=[1, 2, 5], obj_values=[2, 10, -2], capacity=4)
    self.assertTrue(firefly_pool.is_best_fly(firefly_pool._pool[1]))
    self.assertFalse(firefly_pool.is_best_fly(firefly_pool._pool[0]))
    self.assertFalse(firefly_pool.is_best_fly(firefly_pool._pool[2]))

  def test_get_next_moving_fly_copy(self):
    firefly_pool = testing.create_fake_populated_firefly_pool(
        x_values=[1, 2, 5], obj_values=[2, 10, -2], capacity=5)
    firefly_pool._last_id = 1
    moving_fly1 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly1.id_, 2)
    moving_fly2 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly2.id_, 0)
    moving_fly3 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly3.id_, 1)

  def test_get_next_moving_fly_copy_after_removing_last_id_fly(self):
    firefly_pool = testing.create_fake_populated_firefly_pool(
        x_values=[1, 2, 5], obj_values=[2, 10, -2], capacity=5)
    firefly_pool._last_id = 1
    # Remove the fly associated with `_last_id` from the pool.
    del firefly_pool._pool[1]
    moving_fly1 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly1.id_, 2)
    moving_fly2 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly2.id_, 0)
    moving_fly3 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly3.id_, 2)

  def test_get_next_moving_fly_copy_after_removing_multiple_flies(self):
    firefly_pool = testing.create_fake_populated_firefly_pool(
        x_values=[1, 2, 5, -1], obj_values=[2, 10, -2, 8], capacity=5)
    firefly_pool._last_id = 3
    # Remove the several flies
    del firefly_pool._pool[0]
    del firefly_pool._pool[2]
    moving_fly1 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly1.id_, 1)
    moving_fly2 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly2.id_, 3)
    moving_fly3 = firefly_pool.get_next_moving_fly_copy()
    self.assertEqual(moving_fly3.id_, 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Compatibility method comparison for protobufs.

Note that using this method can be unstable and hard to debug.
Provided here for compatibility.
"""

# TODO: Add pytypes.
# TODO: Fix this library.


# pylint: disable=invalid-name
def assertProto2Equal(self, x, y):
  self.assertEqual(x.SerializeToString(), y.SerializeToString())


def assertProto2Contains(self, x, y):
  del self
  if not isinstance(x, str):
    x = str(x)
  if x not in str(y):
    pass


def assertProto2SameElements(*args, **kwargs):
  del args, kwargs
  pass



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for random."""

import collections
import math

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


class RandomTest(absltest.TestCase):

  def test_on_flat_space(self):
    config = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    designer = random.RandomDesigner(config.search_space, seed=None)
    self.assertLen(
        test_runners.run_with_random_metrics(
            designer, config, iters=50, batch_size=1), 50)

  def test_reproducible_random(self):
    config = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    designer = random.RandomDesigner(config.search_space, seed=5)
    t1 = designer.suggest(10)

    designer = random.RandomDesigner(config.search_space, seed=5)
    t2 = designer.suggest(10)
    self.assertEqual(t1, t2)

  def test_log_scaling(self):
    """Confirm that LINEAR and LOG scaling give the right distribution."""
    config = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    designer = random.RandomDesigner(config.search_space, seed=None)
    trials = test_runners.run_with_random_metrics(
        designer, config, iters=1000, batch_size=1
    )
    # LINEAR scaling.
    sum_lineardouble = 0.0
    for t in trials:
      sum_lineardouble += t.parameters['lineardouble'].value
    avg_lineardouble = sum_lineardouble / len(trials)
    # Delta=0.16 corresponds to ~6 standard deviations, so the probability of
    # accidental test failure is ~1e-6.
    self.assertAlmostEqual(avg_lineardouble, 0.5 * (-1 + 2), delta=0.16)
    # LOG scaling
    sum_log_logdouble = 0.0
    sum_logdouble = 0.0
    for t in trials:
      sum_logdouble += t.parameters['logdouble'].value
      sum_log_logdouble += math.log(t.parameters['logdouble'].value)
    avg_logdouble = sum_logdouble / len(trials)
    avg_log_logdouble = sum_log_logdouble / len(trials)
    # If the distribution were LINEAR, we'd expect avg_logdouble = 50;
    # if the distribution were LOG, we'd expect avg_logdouble = 7.2, so
    # we check that avg_logdouble is closer to the LOG expected value.
    self.assertLess(abs(avg_logdouble - 7.2), abs(avg_logdouble - 50))
    # The average of log(value) should be close to the average of
    # the logs of the endpoints.  $delta is set at 6 standard deviations to
    # yield a false failure rate of about 1e-6.
    self.assertAlmostEqual(
        avg_log_logdouble, 0.5 * (math.log(1e-4) + math.log(1e2)), delta=0.76
    )

  def test_discrete_parameter(self):
    """Confirm that DISCRETE parameters are sampled uniformally."""
    n_samples = 10_000
    search_space = vz.SearchSpace()
    search_space.root.add_discrete_param(
        'discrete1', feasible_values=[0, 70, 71, 72, 100]
    )
    search_space.root.add_discrete_param('discrete2', feasible_values=[0, 8, 9])
    random_designer = random.RandomDesigner(search_space, seed=1)
    values = [
        (sg.parameters['discrete1'].value, sg.parameters['discrete2'].value)
        for sg in random_designer.suggest(n_samples)
    ]
    counts = collections.defaultdict(int)
    for value in values:
      counts[str(value[0]) + str(value[1])] += 1
    for count in counts.values():
      # Check that there's no more than 1% deviation.
      self.assertLess(abs(count / n_samples - 1 / 15), 0.01)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""HPOB Handler copied (with slightly modified filesystem logic) from

https://github.com/releaunifreiburg/HPO-B/blob/main/hpob_handler.py.
"""
# pylint:skip-file

import json
import os

import functools
import numpy as np
import xgboost as xgb

Open = open
Exists = os.path.exists
IsDir = os.path.isdir


class HPOBHandler:

  def __init__(self,
               root_dir="hpob-data/",
               mode="v3-test",
               surrogates_dir="saved-surrogates/"):
    """Constructor for the HPOBHandler.

    Inputs:
        * root_dir: path to directory with the benchmark data.
        * mode: mode name indicating how to load the data. Options:
            - v1: Loads HPO-B-v1
            - v2: Loads HPO-B-v2
            - v3: Loads HPO-B-v3
            - v3-test: Loads only the meta-test split from HPO-B-v3
            - v3-train-augmented: Loads all splits from HPO-B-v3, but
            augmenting the meta-train data with the less frequent
            search-spaces.
        * surrogates_dir: path to directory with surrogates models.
    """

    print("Loading HPO-B handler")
    self.mode = mode
    self.surrogates_dir = surrogates_dir
    self.seeds = ["test0", "test1", "test2", "test3", "test4"]

    if self.mode == "v3-test":
      self.load_data(root_dir, only_test=True)
    elif self.mode == "v3-train-augmented":
      self.load_data(root_dir, only_test=False, augmented_train=True)
    elif self.mode in ["v1", "v2", "v3"]:
      self.load_data(root_dir, version=self.mode, only_test=False)
    else:
      raise ValueError("Provide a valid mode")

    self.surrogates_stats = HPOBHandler._cached_surrogates_stats(surrogates_dir)

  @classmethod
  @functools.lru_cache(maxsize=128)
  def _cached_load_data(cls,
                        rootdir="",
                        version="v3",
                        only_test=True,
                        augmented_train=False):

    meta_train_data = {}
    meta_validation_data = {}
    meta_test_data = {}

    print("Loading data...")
    meta_train_augmented_path = os.path.join(
        rootdir, "meta-train-dataset-augmented.json")
    meta_train_path = os.path.join(rootdir, "meta-train-dataset.json")
    meta_test_path = os.path.join(rootdir, "meta-test-dataset.json")
    meta_validation_path = os.path.join(rootdir, "meta-validation-dataset.json")
    bo_initializations_path = os.path.join(rootdir, "bo-initializations.json")

    with Open(meta_test_path, "rb") as f:
      meta_test_data = json.load(f)

    with Open(bo_initializations_path, "rb") as f:
      bo_initializations = json.load(f)

    if not only_test:
      if augmented_train or version == "v1":
        with Open(meta_train_augmented_path, "rb") as f:
          meta_train_data = json.load(f)
      else:
        with Open(meta_train_path, "rb") as f:
          meta_train_data = json.load(f)
      with Open(meta_validation_path, "rb") as f:
        meta_validation_data = json.load(f)

    if version != "v3":
      temp_data = {}
      for search_space in meta_train_data.keys():
        temp_data[search_space] = {}

        for dataset in meta_train_data[search_space].keys():
          temp_data[search_space][dataset] = meta_train_data[search_space][
              dataset]

        if search_space in meta_test_data.keys():
          for dataset in meta_test_data[search_space].keys():
            temp_data[search_space][dataset] = meta_test_data[search_space][
                dataset]

          for dataset in meta_validation_data[search_space].keys():
            temp_data[search_space][dataset] = meta_validation_data[
                search_space][dataset]
      meta_test_data = temp_data

    return meta_train_data, meta_validation_data, meta_test_data, bo_initializations

  @classmethod
  @functools.lru_cache(maxsize=128)
  def _cached_surrogates_stats(cls, surrogates_dir: str):
    surrogates_file = surrogates_dir + "summary-stats.json"
    if (Exists(surrogates_file) and not IsDir(surrogates_file)):
      with Open(surrogates_file, "rt") as f:
        surrogates_stats = json.load(f)
    else:
      surrogates_stats = None
    return surrogates_stats

  def load_data(self,
                rootdir="",
                version="v3",
                only_test=True,
                augmented_train=False):
    """
        Loads data with some specifications.
        Inputs:
            * root_dir: path to directory with the benchmark data.
            * version: name indicating what HPOB version to use. Options: v1,
            v2, v3).
            * Only test: Whether to load only testing data (valid only for
            version v3).  Options: True/False
            * augmented_train: Whether to load the augmented train data (valid
            only for version v3). Options: True/False
    """

    self.meta_train_data, self.meta_validation_data, self.meta_test_data, self.bo_initializations = HPOBHandler._cached_load_data(
        rootdir=rootdir,
        version=version,
        only_test=only_test,
        augmented_train=augmented_train)

  def normalize(self, y, y_min=None, y_max=None):

    if y_min is None:
      return (y - np.min(y)) / (np.max(y) - np.min(y))
    else:
      return (y - y_min) / (y_max - y_min)

  def evaluate(self,
               bo_method=None,
               search_space_id=None,
               dataset_id=None,
               seed=None,
               n_trials=10):
    """Evaluates a method on the benchmark with discretized search-spaces.

    Inputs:
        * bo_method: object to evaluate. It should have a function (class
        method) named 'observe_and_suggest'.
        * search_space_id: Identifier of the search spaces for the
        evaluation. Option: see original paper.
        * dataset_id: Identifier of the dataset for the evaluation. Options:
        see original paper.
        * seed: Identifier of the seed for the evaluation. Options: test0,
        test1, test2, test3, test4.
        * trails: Number of trials (iterations on the opoitmization).
    Ooutput:
        * a list with the maximumu performance (incumbent) for every trial.
    """

    assert bo_method != None, "Provide a valid method object for evaluation."
    assert hasattr(bo_method, "observe_and_suggest"), (
        "The provided  object does not have a method called "
        "´observe_and_suggest´")
    assert search_space_id != None, ("Provide a valid search space id. See "
                                     "documentatio for valid obptions.")
    assert dataset_id != None, ("Provide a valid dataset_id. See documentation "
                                "for valid options.")
    assert seed != None, ("Provide a valid initialization. Valid options are: "
                          "test0, test1, test2, test3, test4.")

    n_initial_evaluations = 5
    X = np.array(self.meta_test_data[search_space_id][dataset_id]["X"])
    y = np.array(self.meta_test_data[search_space_id][dataset_id]["y"])
    y = self.normalize(y)
    data_size = len(X)

    pending_evaluations = list(range(data_size))
    current_evaluations = []

    init_ids = self.bo_initializations[search_space_id][dataset_id][seed]

    for i in range(n_initial_evaluations):
      idx = init_ids[i]
      pending_evaluations.remove(idx)
      current_evaluations.append(idx)

    max_accuracy_history = [np.max(y[current_evaluations])]
    for i in range(n_trials):

      idx = bo_method.observe_and_suggest(X[current_evaluations],
                                          y[current_evaluations],
                                          X[pending_evaluations])
      idx = pending_evaluations[idx]
      pending_evaluations.remove(idx)
      current_evaluations.append(idx)
      max_accuracy_history.append(np.max(y[current_evaluations]))

    return max_accuracy_history

  def evaluate_continuous(self,
                          bo_method=None,
                          search_space_id=None,
                          dataset_id=None,
                          seed=None,
                          n_trials=10):
    """
        Evaluates a method on the benchmark with continuous search-spaces.
        Inputs:
            * bo_method: object to evaluate. It should have a function (class
            method) named 'observe_and_suggest'.
            * search_space_id: Identifier of the search spaces for the
            evaluation. Option: see original paper.
            * dataset_id: Identifier of the dataset for the evaluation. Options:
            see original paper.
            * seed: Identifier of the seed for the evaluation. Options: test0,
            test1, test2, test3, test4.
            * trails: Number of trials (iterations on the opoitmization).
        Ooutput:
            * a list with the maximumu performance (incumbent) for every trial.
        """

    assert bo_method != None, "Provide a valid method object for evaluation."
    assert hasattr(bo_method, "observe_and_suggest"), (
        "The provided  object does not have a method called "
        "´observe_and_suggest´")
    assert search_space_id != None, ("Provide a valid search space id. See "
                                     "documentatio for valid obptions.")
    assert dataset_id != None, ("Provide a valid dataset_id. See documentation "
                                "for valid options.")
    assert seed != None, ("Provide a valid initialization. Valid options are: "
                          "test0, test1, test2, test3, test4.")

    surrogate_name = "surrogate-" + search_space_id + "-" + dataset_id
    bst_surrogate = xgb.Booster()
    bst_surrogate.load_model(self.surrogates_dir + surrogate_name + ".json")

    n_initial_evaluations = 5
    X = np.array(self.meta_test_data[search_space_id][dataset_id]["X"])
    y = np.array(self.meta_test_data[search_space_id][dataset_id]["y"])
    y_min = self.surrogates_stats[surrogate_name]["y_min"]
    y_max = self.surrogates_stats[surrogate_name]["y_max"]
    dim = X.shape[1]
    current_evaluations = []
    init_ids = self.bo_initializations[search_space_id][dataset_id][seed]

    for i in range(n_initial_evaluations):
      idx = init_ids[i]
      current_evaluations.append(idx)

    x_observed = X[current_evaluations]
    y_observed = y[current_evaluations]

    max_accuracy_history = []

    for i in range(n_trials):
      y_tf_observed = self.normalize(y_observed, y_min, y_max)
      y_tf_observed = np.clip(y_tf_observed, 0, 1)
      best_f = np.max(y_tf_observed)
      max_accuracy_history.append(best_f)

      new_x = bo_method.observe_and_suggest(x_observed, y_tf_observed)
      x_q = xgb.DMatrix(new_x.reshape(-1, dim))
      new_y = bst_surrogate.predict(x_q)

      y_observed = np.append(y_observed, new_y).reshape(-1, 1)
      x_observed = np.append(x_observed, new_x).reshape(-1, x_observed.shape[1])

    y_tf_observed = self.normalize(y_observed, y_min, y_max)
    y_tf_observed = np.clip(y_tf_observed, 0, 1)
    max_accuracy_history.append(best_f)

    return max_accuracy_history

  def get_seeds(self):
    return self.seeds



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for pareto_optimal."""

import numpy as np

from vizier._src.pyvizier.multimetric import pareto_optimal
from absl.testing import absltest


class ParetoOptimalTest(absltest.TestCase):

  def setUp(self):
    super(ParetoOptimalTest, self).setUp()
    self.algo = pareto_optimal.FastParetoOptimalAlgorithm(
        base_algorithm=pareto_optimal.NaiveParetoOptimalAlgorithm())

  def test_is_pareto_optimal(self):
    points = np.array([[1, 2, 3], [1, 2, 3], [2, 4, 1], [1, 2, -1]])
    np.testing.assert_array_equal(
        self.algo.is_pareto_optimal(points), [True, True, True, False])

    # Adding a point that dominates the first two values.
    points = np.vstack([points, [2, 3, 4]])
    np.testing.assert_array_equal(
        self.algo.is_pareto_optimal(points), [False, False, True, False, True])

  def test_is_pareto_optimal_all_optimal(self):
    # Generate points on positive orthant sphere.
    points = abs(np.random.normal(size=(1000, 3)))
    points /= np.linalg.norm(points, axis=1)[..., np.newaxis]
    algo = pareto_optimal.FastParetoOptimalAlgorithm(
        base_algorithm=pareto_optimal.NaiveParetoOptimalAlgorithm(),
        recursive_threshold=100)
    self.assertTrue(np.all(algo.is_pareto_optimal(points)))

  def test_is_pareto_optimal_randomized(self):
    dim = 10
    points = np.random.normal(size=(1000, dim))
    # Make a copy of points to test equality case.
    points = np.vstack([points, points])

    # Make sure the fast and base algorithms match.
    simple_test = self.algo.is_pareto_optimal(points)
    fast_test = self.algo._base_algorithm.is_pareto_optimal(points)

    self.assertTrue(np.all(simple_test == fast_test))

  def test_is_pareto_optimal_against(self):
    points = np.array([[1, 2, 3], [2, 4, 1], [1, 2, -1]])
    dominating_points = np.array([[1, 2, 3], [3, 4, 0]])

    np.testing.assert_array_equal(
        self.algo.is_pareto_optimal_against(
            points, dominating_points, strict=True), [True, True, False])
    np.testing.assert_array_equal(
        self.algo.is_pareto_optimal_against(
            points, dominating_points, strict=False), [False, True, False])

  def test_is_pareto_optimal_against_randomized(self):
    dim = 4
    dominating_points = np.random.normal(size=(10000, dim))
    points = np.random.normal(size=(1000, dim))

    # Make sure the fast and base algorithms match.
    simple_check = self.algo.is_pareto_optimal_against(
        points, dominating_points, strict=True)
    fast_check = self.algo._base_algorithm.is_pareto_optimal_against(
        points, dominating_points, strict=True)

    # Results should not be affected by changing recursive threshold.
    self.assertTrue(np.all(simple_check == fast_check))


if __name__ == '__main__':
  absltest.main()
